import os
import sys
import argparse
import json

import gdown

from idtrackerai.constants import IDTRACKERAI_FOLDER, TEST_VIDEO_URL

def create_data_folder():
    idtrackerai_data_folder = os.path.join(os.path.dirname(IDTRACKERAI_FOLDER), 'data')
    if not os.path.isdir(idtrackerai_data_folder):
        os.mkdir(idtrackerai_data_folder)

    return idtrackerai_data_folder


def download_example_video(output_folder=''):
    if not output_folder:
        output_folder = create_data_folder()

    video_example_folder = os.path.join(output_folder, 'idtrackerai_example_video')
    if not os.path.isdir(video_example_folder):
        os.mkdir(video_example_folder)

    output_file_avi = os.path.join(video_example_folder, 'example_video_idtrackerai.avi')

    gdown.download(TEST_VIDEO_URL, output_file_avi, quiet=False)

    return output_file_avi


def update_json(video_path, args):
    json_file_path = os.path.join(IDTRACKERAI_FOLDER, 'tests', 'test.json')

    with open(json_file_path) as json_file:
        json_content = json.load(json_file)

    json_content['_video']['value'] = video_path
    if args.no_identities:
        json_content['_no_ids']['value'] = "True"

    with open(json_file_path, 'w') as json_file:
        json.dump(json_content, json_file)

    return json_file_path


def test():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output_folder",
                        type=str,
                        help="Path to the folder where the video will be stored")
    parser.add_argument("-n", "--no_identities",
                        action='store_true',
                        help="Flag to track without identities")
    args = parser.parse_args()
    video_path = download_example_video(args.output_folder)
    json_file_path = update_json(video_path, args)
    os.system('idtrackerai terminal_mode --load {} --exec track_video'.format(json_file_path))


if __name__ == "__main__":
    test()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import numpy as np
from pprint import pprint
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.blob import Blob
from idtrackerai.groundtruth_utils.generate_groundtruth import GroundTruth, GroundTruthBlob

"""Given two list of blobs_in_video, one deduced from human groundtruth and the other
generated by the tracking algorithm, compares them and gives back some statistics

Crossing: crossings are a special case. We ...
"""

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.compute_statistics_against_groundtruth")


def get_corresponding_gt_blob(blob, gt_blobs_in_frame):
    corresponding_gt_blobs = []
    for gt_blob in gt_blobs_in_frame:
        num_overlapping_pixels = len(set(gt_blob.pixels) & set(blob.pixels))
        if num_overlapping_pixels > 0:
            corresponding_gt_blobs.append(gt_blob)
    return corresponding_gt_blobs


def compare_tracking_against_groundtruth(number_of_animals,
                                         blobs_in_video_groundtruth,
                                         blobs_in_video,
                                         identities_dictionary_permutation):
    # create dictionary to store eventual corrections made by the user
    results = {}
    results['number_of_blobs_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['sum_individual_P2'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_assigned_blobs_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_blobs_assigned_during_accumulation_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_blobs_after_accumulation_per_identity'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_all_blobs'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_assigned_blobs'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_assigned_during_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_after_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_errors_in_blobs_assigned_after_accumulation'] = {i:0 for i in range(1, number_of_animals + 1)}
    results['number_of_individual_blobs'] = 0
    results['number_of_crossing_blobs'] = 0
    results['number_of_crossings_blobs_assigned_as_individuals'] = 0
    results['frames_with_identity_errors'] = []
    results['fragment_identifiers_with_identity_errors'] = []
    results['frames_with_crossing_errors'] = []
    results['fragment_identifiers_with_crossing_errors'] = []
    results['frames_with_zeros_in_groundtruth'] = []
    results['number_of_crossing_fragments'] = 0
    results['fragments_identifiers_of_crossings'] = []

    for gt_blobs_in_frame, blobs_in_frame in zip(blobs_in_video_groundtruth, blobs_in_video):
        for blob in blobs_in_frame:
            corresponding_gt_blobs = get_corresponding_gt_blob(blob, gt_blobs_in_frame)
            if blob.is_an_individual and len(corresponding_gt_blobs) == 1:
                groundtruth_blob = corresponding_gt_blobs[0]
                if groundtruth_blob.is_an_individual and groundtruth_blob.identity != -1 and not groundtruth_blob.was_a_crossing: # we are not considering crossing or failures of the model area
                    if identities_dictionary_permutation is not None:
                        gt_identity = identities_dictionary_permutation[groundtruth_blob.identity]
                    else:
                        gt_identity = groundtruth_blob.identity
                    results['number_of_individual_blobs'] += 1
                    if gt_identity == 0:
                        results['frames_with_zeros_in_groundtruth'].append(groundtruth_blob.frame_number)
                    else:
                        try:
                            if blob.assigned_identity != 0 and blob.identity_corrected_closing_gaps is None: # we only consider P2 for non interpolated blobs
                                results['sum_individual_P2'][gt_identity] += blob._P2_vector[gt_identity - 1]
                        except:
                            logger.debug("P2_vector %s" %str(blob._P2_vector))
                            logger.debug("individual %s" %str(blob.is_an_individual))
                            logger.debug("fragment identifier ", blob.fragment_identifier)
                        results['number_of_blobs_per_identity'][gt_identity] += 1
                        results['number_of_assigned_blobs_per_identity'][gt_identity] += 1 if blob.assigned_identity != 0 else 0
                        results['number_of_blobs_assigned_during_accumulation_per_identity'][gt_identity] += 1 if blob.used_for_training else 0
                        results['number_of_blobs_after_accumulation_per_identity'][gt_identity] += 1 if not blob.used_for_training else 0
                        if gt_identity != blob.assigned_identity:
                            results['number_of_errors_in_all_blobs'][gt_identity] += 1
                            results['number_of_errors_in_blobs_after_accumulation'][gt_identity] += 1 if not blob.used_for_training else 0
                            if blob.assigned_identity != 0:
                                results['number_of_errors_in_assigned_blobs'][gt_identity] += 1
                                results['number_of_errors_in_blobs_assigned_during_accumulation'][gt_identity] += 1 if blob.used_for_training else 0
                                results['number_of_errors_in_blobs_assigned_after_accumulation'][gt_identity] += 1 if not blob.used_for_training else 0
                            if blob.fragment_identifier not in results['fragment_identifiers_with_identity_errors']:
                                results['frames_with_identity_errors'].append(blob.frame_number)
                                results['fragment_identifiers_with_identity_errors'].append(blob.fragment_identifier)

                elif groundtruth_blob.is_a_crossing or gt_identity == -1:
                    if blob.fragment_identifier not in results['fragments_identifiers_of_crossings']:
                        results['fragments_identifiers_of_crossings'].append(blob.fragment_identifier)
                        results['number_of_crossing_fragments'] += 1
                    results['number_of_crossing_blobs'] += 1
                    results['number_of_crossings_blobs_assigned_as_individuals'] += 1 if blob.is_an_individual else 0
                    if blob.is_an_individual:
                        if blob.fragment_identifier not in results['fragment_identifiers_with_crossing_errors']:
                            results['frames_with_crossing_errors'].append(blob.frame_number)
                            results['fragment_identifiers_with_crossing_errors'].append(blob.fragment_identifier)

    return results


def check_ground_truth_consistency(video, gt_video):
    if video.number_of_frames != gt_video.number_of_frames:
        raise ValueError('The number of frames in the video and in the \
                         groundtruth video are different. The groundtruth\
                         file cannot be reused')

def check_first_frame_first_global_fragment(video, first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video):
    while len(blobs_in_video_groundtruth[first_frame_first_global_fragment]) < video.number_of_animals:
        first_frame_first_global_fragment += 1
    return first_frame_first_global_fragment

# def get_permutation_of_identities(video,
#                                   first_frame_first_global_fragment,
#                                   blobs_in_video_groundtruth,
#                                   blobs_in_video):
#     print(first_frame_first_global_fragment)
#     first_frame_first_global_fragment = check_first_frame_first_global_fragment(video, first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video)
#     print(first_frame_first_global_fragment)
#     print(len(blobs_in_video_groundtruth[first_frame_first_global_fragment]))
#     if first_frame_first_global_fragment is not None:
#         groundtruth_identities_in_first_frame = \
#             [blob.identity for blob in
#              blobs_in_video_groundtruth[first_frame_first_global_fragment]]
#         identities_in_first_frame = \
#             [blob.identity for blob in
#              blobs_in_video[first_frame_first_global_fragment]]
#         logger.debug('groundtruth identities in first frame %s'
#                      % str(groundtruth_identities_in_first_frame))
#         logger.debug('identities in first frame %s'
#                      % str(identities_in_first_frame))
#         identities_dictionary_permutation = \
#             {groundtruth_identity: identity for identity, groundtruth_identity
#              in zip(identities_in_first_frame,
#                     groundtruth_identities_in_first_frame)}
#     else:
#         identities_dictionary_permutation = None
#     print("identities_dictionary_permutation:",
#           identities_dictionary_permutation)
#
#     return identities_dictionary_permutation


def get_permutation_of_identities(video,
                                  first_frame_first_global_fragment,
                                  blobs_in_video_groundtruth,
                                  blobs_in_video):
    first_frame_first_global_fragment = \
        check_first_frame_first_global_fragment(video,
                                                first_frame_first_global_fragment,
                                                blobs_in_video_groundtruth,
                                                blobs_in_video)

    matching_found = False
    while not matching_found:
        gt_blobs_in_first_frame = blobs_in_video_groundtruth[first_frame_first_global_fragment]
        identities_dictionary_permutation = {}
        print(first_frame_first_global_fragment)
        print(len(blobs_in_video[first_frame_first_global_fragment]))
        print(len(blobs_in_video_groundtruth[first_frame_first_global_fragment]))
        for blob in blobs_in_video[first_frame_first_global_fragment]:
            corresponding_blobs = get_corresponding_gt_blob(blob, gt_blobs_in_first_frame)
            if len(corresponding_blobs) == 1:
                identities_dictionary_permutation[corresponding_blobs[0].identity] = blob.identity
            else:
                first_frame_first_global_fragment += 1
                break
        if len(identities_dictionary_permutation)==video.number_of_animals:
            matching_found = True
    print(identities_dictionary_permutation)
    return identities_dictionary_permutation


def get_accuracy_wrt_groundtruth(video, gt_video, blobs_in_video_groundtruth,
                                 blobs_in_video=None,
                                 identities_dictionary_permutation=None):
    number_of_animals = video.number_of_animals
    blobs_in_video = blobs_in_video_groundtruth if blobs_in_video is None\
        else blobs_in_video
    results = compare_tracking_against_groundtruth(number_of_animals,
                                                   blobs_in_video_groundtruth,
                                                   blobs_in_video,
                                                   identities_dictionary_permutation)
    pprint(results)
    if len(results['frames_with_zeros_in_groundtruth']) == 0:
        accuracies = {}
        accuracies['percentage_of_unoccluded_images'] = results['number_of_individual_blobs'] / (results['number_of_individual_blobs'] + results['number_of_crossing_blobs'])
        accuracies['individual_P2_in_validated_part'] = {i : results['sum_individual_P2'][i] / results['number_of_blobs_per_identity'][i]
                                for i in range(1, number_of_animals + 1)}
        accuracies['mean_individual_P2_in_validated_part'] = np.sum(results['sum_individual_P2'].values()) / np.sum(results['number_of_blobs_per_identity'].values())
        accuracies['individual_accuracy'] = {i : 1 - results['number_of_errors_in_all_blobs'][i] / results['number_of_blobs_per_identity'][i]
                                if results['number_of_blobs_per_identity'] != 0 else None
                                for i in range(1, number_of_animals + 1)}
        accuracies['accuracy'] = 1. - np.sum(results['number_of_errors_in_all_blobs'].values()) / np.sum(results['number_of_blobs_per_identity'].values())
        accuracies['individual_accuracy_assigned'] = {i : 1 - results['number_of_errors_in_assigned_blobs'][i] / results['number_of_assigned_blobs_per_identity'][i]
                                        if results['number_of_assigned_blobs_per_identity'] != 0 else None
                                        for i in range(1, number_of_animals + 1)}
        accuracies['accuracy_assigned'] = 1. - np.sum(results['number_of_errors_in_assigned_blobs'].values()) / np.sum(results['number_of_assigned_blobs_per_identity'].values())
        # accuracies['individual_accuracy_in_accumulation'] = {i:1 - results['number_of_errors_in_blobs_assigned_during_accumulation'][i]/results['number_of_blobs_assigned_during_accumulation_per_identity'][i]
        #                         if results['number_of_blobs_assigned_during_accumulation_per_identity'] != 0 else None
        #                         for i in range(1, number_of_animals + 1)}
        accuracies['individual_accuracy_in_accumulation'] = {}
        for i in range(1, number_of_animals + 1):
            if results['number_of_blobs_assigned_during_accumulation_per_identity'][i] != 0:
                accuracies['individual_accuracy_in_accumulation'][i] = 1 - results['number_of_errors_in_blobs_assigned_during_accumulation'][i] / results['number_of_blobs_assigned_during_accumulation_per_identity'][i]
            else:
                accuracies['individual_accuracy_in_accumulation'][i] = None
        accuracies['accuracy_in_accumulation'] = 1. - np.sum(results['number_of_errors_in_blobs_assigned_during_accumulation'].values()) / np.sum(results['number_of_blobs_assigned_during_accumulation_per_identity'].values())
        accuracies['individual_accuracy_after_accumulation'] = {}
        for i in range(1, number_of_animals + 1):
            if results['number_of_blobs_after_accumulation_per_identity'][i] != 0:
                accuracies['individual_accuracy_after_accumulation'][i] = 1 - results['number_of_errors_in_blobs_after_accumulation'][i] / results['number_of_blobs_after_accumulation_per_identity'][i]
            else:
                accuracies['individual_accuracy_after_accumulation'][i] = None
        if np.sum(results['number_of_blobs_after_accumulation_per_identity'].values()) != 0:
            accuracies['accuracy_after_accumulation'] = 1. - np.sum(results['number_of_errors_in_blobs_after_accumulation'].values()) / np.sum(results['number_of_blobs_after_accumulation_per_identity'].values())
        else:
            accuracies['accuracy_after_accumulation'] = None
        if results['number_of_crossing_blobs'] != 0:
            accuracies['crossing_detector_accuracy'] = 1. - results['number_of_crossings_blobs_assigned_as_individuals'] / results['number_of_crossing_blobs']
        else:
            accuracies['crossing_detector_accuracy'] = None
        logger.info("accuracies %s" %str(accuracies))
        logger.info("number of crossing fragments in ground truth interval: %i" %results['number_of_crossing_fragments'])
        logger.info("number of crossing blobs in ground truth interval: %i" %results['number_of_crossing_blobs'])
        return accuracies, results

    else:
        logger.info("there are fish with 0 identity in frame %s" %str(results['frames_with_zeros_in_groundtruth']))
        return None, results


def reduce_pixels(blob, original_width, original_height, width, height, resolution_reduction):
    pxs = np.array(np.unravel_index(blob.pixels, (original_height, original_width))).T
    pxs_reduced = (np.round(pxs*resolution_reduction)).astype('int')
    pxs_reduced = np.ravel_multi_index([pxs_reduced[:, 0], pxs_reduced[:, 1]],
                                       (height, width))
    return pxs_reduced


def reduce_resolution_groundtruth_blobs(groundtruth, video,
                                        blobs_in_video_groundtruth):
    gt_width, gt_height = groundtruth.video.width, groundtruth.video.height
    resolution_reduction = video.resolution_reduction
    for blobs_in_frame in blobs_in_video_groundtruth:
        for blob in blobs_in_frame:
            blob.pixels = reduce_pixels(blob,
                                        video.original_width,
                                        video.original_height,
                                        video.width, video.height,
                                        video.resolution_reduction)


def compute_and_save_session_accuracy_wrt_groundtruth(video,
                                                      groundtruth_type=None):
    logger.info("loading list_of_blobs")
    if groundtruth_type == 'normal':
        list_of_blobs = ListOfBlobs.load(video, video.blobs_path)
    elif groundtruth_type == 'no_gaps':
        list_of_blobs = ListOfBlobs.load(video, video.blobs_no_gaps_path)
    # select ground truth file
    logger.info("loading groundtruth")
    if groundtruth_type == 'normal':
        groundtruth_path = os.path.join(video.video_folder, '_groundtruth.npy')
    elif groundtruth_type == 'no_gaps':
        groundtruth_path = \
            os.path.join(video.video_folder,
                         '_groundtruth_with_crossing_identified.npy')
    groundtruth = np.load(groundtruth_path, allow_pickle=True).item()

    check_ground_truth_consistency(video, groundtruth.video)
    accumulation_number = int(video.accumulation_folder[-1])
    if video.resolution_reduction != 1:
        reduce_resolution_groundtruth_blobs(groundtruth, video,
                                            groundtruth.blobs_in_video)
    identities_dictionary_permutation = \
        get_permutation_of_identities(video,
                                      video.first_frame_first_global_fragment[accumulation_number],
                                      groundtruth.blobs_in_video,
                                      list_of_blobs.blobs_in_video)
    blobs_in_video_groundtruth = \
        groundtruth.blobs_in_video[groundtruth.start:groundtruth.end]
    blobs_in_video = \
        list_of_blobs.blobs_in_video[groundtruth.start:groundtruth.end]
    logger.info("computing groundtruth")
    if groundtruth_type == 'normal':
        accuracies, results = \
            get_accuracy_wrt_groundtruth(video, groundtruth.video,
                                         blobs_in_video_groundtruth,
                                         blobs_in_video,
                                         identities_dictionary_permutation)
    elif groundtruth_type == 'no_gaps':
        accuracies, results = \
            get_accuracy_wrt_groundtruth_no_gaps(video, groundtruth,
                                                 blobs_in_video_groundtruth,
                                                 blobs_in_video,
                                                 identities_dictionary_permutation)
    if accuracies is not None:
        logger.info("saving accuracies in video")
        video.gt_start_end = (groundtruth.start, groundtruth.end)
        if groundtruth_type == 'normal':
            video.gt_accuracy = accuracies
            video.gt_results = results
        elif groundtruth_type == 'no_gaps':
            video.gt_accuracy_no_gaps = accuracies
            video.gt_results_no_gaps = results
        video.save()
    return video, groundtruth


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-gt", "--groundtruth_type", type=str,
                        help="type of groundtruth to compute \
                        ('no_gaps' or 'normal')")
    parser.add_argument("-sf", "--session_folder", type=str,
                        help="path to the session folder")
    args = parser.parse_args()

    groundtruth_type = args.groundtruth_type
    session_folder = args.session_folder
    video_object_path = os.path.join(session_folder, 'video_object.npy')
    logger.info("loading video object")
    video = np.load(video_object_path, allow_pickle=True).item(0)
    video.update_paths(video_object_path)
    groundtruth_path = os.path.join(video.video_folder, '_groundtruth.npy')
    groundtruth = np.load(groundtruth_path, allow_pickle=True).item()
    video, groundtruth = compute_and_save_session_accuracy_wrt_groundtruth(video, groundtruth_type)
    print(video.gt_accuracy)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

from confapp import conf

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.network_params")

class NetworkParams(object):
    """Manages the network hyperparameters and other variables related to the
    identification model (see :class:`~idCNN`)

    Attributes
    ----------
    video_path : string
        Path to the video file
    number_of_animals : int
        Number of animals in the video
    learning_rate : float
        Learning rate for the optimizer
    keep_prob : float
        Dropout probability
    _restore_folder : string
        Path to the folder where the model to be restored is
    _save_folder : string
        Path to the folder where the checkpoints of the current model are stored
    _knowledge_transfer_folder : string
        Path to the folder where the model to be used for knowledge transfer is saved
    use_adam_optimiser : bool
        Flag indicating to use the Adam optimizer with the parameters indicated in _[2]
    scopes_layers_to_optimize : list
        List with the scope names of the layers to be optimized
    _cnn_model : int
        Number indicating the model number to be used from the dictionary of models
        CNN_MODELS_DICT in :mod:`id_CNN`
    image_size : tuple
        Tuple (height, width, channels) for the input images
    number_of_channels : int
        Number of channels of the input image
    kt_conv_layers_to_discard : str
        convolutional layers to discard when performing knowledge transfer

    .. [2] Kingma, Diederik P., and Jimmy Ba. "Adam: A method for stochastic optimization." arXiv preprint arXiv:1412.6980 (2014).
    """

    def __init__(self, number_of_animals, cnn_model = conf.CNN_MODEL, learning_rate = None,
                keep_prob = None,
                use_adam_optimiser = conf.USE_ADAM_OPTIMISER,
                scopes_layers_to_optimize = None, restore_folder = None,
                save_folder = None, knowledge_transfer_folder = conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN,
                image_size = None,
                number_of_channels = None,
                video_path = None):
        self.video_path = video_path
        self.number_of_animals = number_of_animals
        self.learning_rate = learning_rate
        self.keep_prob = keep_prob
        self._restore_folder = restore_folder
        self._save_folder = save_folder
        self._knowledge_transfer_folder = knowledge_transfer_folder
        self.use_adam_optimiser = use_adam_optimiser
        self.scopes_layers_to_optimize = scopes_layers_to_optimize
        self._cnn_model = cnn_model
        self.image_size = image_size
        self.target_image_size = None
        self.pre_target_image_size = None
        self.action_on_image = None
        self.number_of_channels = number_of_channels
        self._kt_conv_layers_to_discard = None

    @property
    def cnn_model(self):
        return self._cnn_model

    @property
    def restore_folder(self):
        return self._restore_folder

    @restore_folder.setter
    def restore_folder(self, path):
        assert os.path.isdir(path)
        self._restore_folder = path

    @property
    def save_folder(self):
        return self._save_folder

    @save_folder.setter
    def save_folder(self, path):
        if not os.path.isdir(path):
            os.path.makedirs(path)
        self._save_folder = path

    @property
    def knowledge_transfer_folder(self):
        return self._knowledge_transfer_folder

    @property
    def kt_conv_layers_to_discard(self):
        return self._kt_conv_layers_to_discard

    @knowledge_transfer_folder.setter
    def knowledge_transfer_folder(self, path):
        assert os.path.isdir(path)
        self._knowledge_transfer_folder = path



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np

from idtrackerai.trainer import train
from idtrackerai.accumulation_manager import get_predictions_of_candidates_fragments
from confapp import conf


if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger('__main__.accumulator')


def perform_one_accumulation_step(accumulation_manager,
                                  video,
                                  global_step,
                                  net,
                                  GUI_axes=None,
                                  net_properties=None,
                                  plot_flag=False,
                                  save_summaries=False):
    logger.info("accumulation step %s" % accumulation_manager.counter)
    video.accumulation_step = accumulation_manager.counter
    accumulation_manager.get_new_images_and_labels()
    images, labels = accumulation_manager.get_images_and_labels_for_training()
    logger.debug("images: %s" % str(images.shape))
    logger.debug("labels: %s" % str(labels.shape))
    global_step, net,\
        store_validation_accuracy_and_loss_data,\
        store_training_accuracy_and_loss_data = \
        train(video,
              accumulation_manager.list_of_fragments.fragments,
              net, images, labels,
              store_accuracy_and_error=True,
              check_for_loss_plateau=True,
              save_summaries=save_summaries,
              print_flag=False,
              plot_flag=plot_flag,
              global_step=global_step,
              accumulation_manager=accumulation_manager)
    if net_properties is not None:
        net_properties.setter(global_step)
    # update the set of images used for training
    logger.info("Update images and labels used for training")
    accumulation_manager.update_used_images_and_labels()
    # assign identities fo the global fragments that have been used for training
    logger.info("Assigning identities to accumulated global fragments")
    accumulation_manager.assign_identities_to_fragments_used_for_training()
    # update the list of individual fragments that have been used for training
    logger.info("Update list of individual fragments used for training")
    accumulation_manager.update_list_of_individual_fragments_used()
    # compute ratio of accumulated images and stop if it is above random
    accumulation_manager.ratio_accumulated_images = accumulation_manager.list_of_fragments.compute_ratio_of_images_used_for_training()
    logger.info("The %f percent of the images has been accumulated" %(accumulation_manager.ratio_accumulated_images * 100))
    if accumulation_manager.ratio_accumulated_images > conf.THRESHOLD_EARLY_STOP_ACCUMULATION:
        logger.debug("Stopping accumulation by early stopping criteria")
        return accumulation_manager.ratio_accumulated_images, store_validation_accuracy_and_loss_data, store_training_accuracy_and_loss_data
    # Set accumulation parameters for rest of the accumulation
    #take images from global fragments not used in training (in the remainder test global fragments)
    logger.info("Get new global fragments for training")
    if any([not global_fragment.used_for_training for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments]):
        logger.info("Generate predictions on candidate global fragments")
        predictions,\
        softmax_probs,\
        indices_to_split,\
        candidate_individual_fragments_identifiers = get_predictions_of_candidates_fragments(net,
                                                                                            video,
                                                                                            accumulation_manager.list_of_fragments.fragments)
        logger.debug('Splitting predictions by fragments...')
        accumulation_manager.split_predictions_after_network_assignment(predictions,
                                                                        softmax_probs,
                                                                        indices_to_split,
                                                                        candidate_individual_fragments_identifiers)
        # assign identities to the global fragments based on the predictions
        logger.info("Checking eligibility criteria and generate the new list of global fragments to accumulate")
        accumulation_manager.get_acceptable_global_fragments_for_training(candidate_individual_fragments_identifiers)
        #Million logs
        logger.info("Number of non certain global fragments: %i" %accumulation_manager.number_of_noncertain_global_fragments)
        logger.info("Number of randomly assigned global fragments: %i" %accumulation_manager.number_of_random_assigned_global_fragments)
        logger.info("Number of non consistent global fragments: %i " %accumulation_manager.number_of_nonconsistent_global_fragments)
        logger.info("Number of non unique global fragments: %i " %accumulation_manager.number_of_nonunique_global_fragments)
        logger.info("Number of acceptable global fragments: %i " %accumulation_manager.number_of_acceptable_global_fragments)
        logger.info("Number of non certain fragments: %i" %accumulation_manager.number_of_noncertain_fragments)
        logger.info("Number of randomly assigned fragments: %i" %accumulation_manager.number_of_random_assigned_fragments)
        logger.info("Number of non consistent fragments: %i " %accumulation_manager.number_of_nonconsistent_fragments)
        logger.info("Number of non unique fragments: %i " %accumulation_manager.number_of_nonunique_fragments)
        logger.info("Number of acceptable fragments: %i " %accumulation_manager.number_of_acceptable_fragments)
        new_values = [len([global_fragment for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments if global_fragment.used_for_training]),
                        accumulation_manager.number_of_noncertain_global_fragments,
                        accumulation_manager.number_of_random_assigned_global_fragments,
                        accumulation_manager.number_of_nonconsistent_global_fragments,
                        accumulation_manager.number_of_nonunique_global_fragments,
                        np.sum([global_fragment.acceptable_for_training(accumulation_manager.accumulation_strategy) for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments]),
                        store_validation_accuracy_and_loss_data.accuracy,
                        store_validation_accuracy_and_loss_data.individual_accuracy,
                        store_training_accuracy_and_loss_data.accuracy,
                        store_training_accuracy_and_loss_data.individual_accuracy,
                        accumulation_manager.ratio_accumulated_images]
        video.store_accumulation_step_statistics_data(new_values)
        accumulation_manager.update_counter()
    # else:
    #     logger.info("All the global fragments have been used for accumulation")
    #     break

    accumulation_manager.ratio_accumulated_images = accumulation_manager.list_of_fragments.compute_ratio_of_images_used_for_training()
    video.store_accumulation_statistics_data(video.accumulation_trial)
    return accumulation_manager.ratio_accumulated_images, store_validation_accuracy_and_loss_data, store_training_accuracy_and_loss_data



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import numpy as np
from confapp import conf

from idtrackerai.utils.py_utils import delete_attributes_from_object

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.fragment")

class Fragment(object):
    """ Collects the Blob objects (:class:`~blob.Blob`) associated to the same individual or crossing.

    Attributes
    ----------

    fragment_identifier : int
        Unique identifier of the fragment assigned with :func:`~list_of_blobs.ListOfBlobs.compute_fragment_identifier_and_blob_index` and :func:`~list_of_blobs.ListOfBlobs.compute_crossing_fragment_identifier`
    start_end : tuple
        (starting_frame,  ending_frame) of the fragment
    blob_hierarchy_in_starting_frame : int
        Hierarchy of the blob in the starting frame of the fragment (the hierarchy if compute bottom-top, left-right).
    images : list
        List of images associated to every blob represented in the fragment and ordered according to the frame they have been segmented from
    bounding_box_in_frame_coordinates : tuple
        List of bounding boxes (see :attr:`~blob.Blob.bounding_box_in_frame_coordinates`)
    centroids : list
        List of centroids (see :attr:`~blob.Blob.centroid`)
    areas : list
        List of areas (see :attr:`~blob.Blob.area`)
    is_an_individual : bool
        True if the image has been classified as representing an individual by :func:`~crossing_detector.detect_crossings`
    is_a_crossing : bool
        True if the image has been classified as representing a crossing by :func:`~crossing_detector.detect_crossings`
    number_of_animals : int
        Number of animal to be tracked
    user_generated_identity : int
        Identity generated by the user during validation (if it exists, else None)
    is_in_a_global_fragment : bool
        True if the fragment is contained in a global fragment (see :class:`~globalfragment.GlobalFragment`)
    used_for_training : bool
        True if the fragment has been used to train the idCNN in the finerprinting protocol used to track the video
    accumulated_globally : bool
        True if the fragment's images have been accumulated with global strategy as references during the fingerprinting protocol used to track the video
    accumulated_partially : bool
        True if the fragment's images have been accumulated with partial strategy as references during the fingerprinting protocol used to track the video
    accumulation_step : int
        Accumulation step in which the fragment's images have been accumulated
    accumulable : bool
        True if the fragment is eligible for accumulation i.e. it is contained in a global fragment object (see :class:`~globalfragment.GlobalFragment`)
    used_for_pretraining : bool
        True if the fragment has been used to pretrain the idCNN in the third protocol (see :func:`~pre_trainer.pre_train`)
    acceptable_for_training : bool
        True if the fragment satisfies the conditions in :meth:`~accumulation_manager.AccumulationManager.check_if_is_acceptable_for_training`
    frequencies : ndarray
        array with shape [1, number_of_animals]. The ith component is the number of times the fragment has been assigned by the idCNN (argmax of the softmax predictions) to the identity i
    P1_vector : ndarray
        array with shape [1, number_of_animals] computed from frequencies by :meth:`compute_identification_statistics`
    P2_vector : ndarray
        array with shape [1, number_of_animals] computed from the P1_vector and considering all the identified fragment overlapping (in frames) with self by :meth:`compute_identification_statistics`
    certainty : float
        certainty of the identification of the fragment computed according to meth:`compute_certainty_of_individual_fragment` (depends only on the predictions of the idCNN for the fragment's images)
    certainty_P2 : float
        certainty of the identification of the fragment computed according to meth:`compute_P2_vector`
    is_certain : bool
        True if :attr:`certainty` is greater than :const:`conf.CERTAINTY_THRESHOLD`
    temporary_id : int
        Identity assigned to the fragment during the fingerprint protocols cascade
    identity : int
        Identity assigned to the fragment
    identity_is_fixed : bool
        True if the :attr:`certainty_P2` is greater than :const:`conf.FIXED_IDENTITY_THRESHOLD`
    identities_corrected_closing_gaps : list
        Identity assigned to the fragment while solving the crossing if it exists else None
    user_generated_identity : int
        Identity assigned to the fragment by the user during validation if it exists else None
    final_identities : list
        Final identities of the fragment. It corresponds to :attr:`user_generated_identities` if it exist otherwise to the identity assigned by the algorithm
    assigned_identities : list
        Identity assigned to the fragment by the algorithm
    ambiguous_identities : list
        List of possible identities in case the assignment is ambiguous (two or more identity can be assigned to the fragment with the same certainty)
    potentially_randomly_assigned : bool
        True if :attr:`certainty` is below random wrt the number of images in the fragment
    non_consistent : bool
        True if exist a fragment that overlaps in frame with self whose identity is the same as self
    number_of_images : int
        number of images composing the fragment
    has_enough_accumulated_coexisting_fragments : bool
        the partial accumulation strategy assumes that the condition of :attr:`has_enough_accumulated_coexisting_fragments` holds
    distance_travelled : float
        distance travelled by the individual in the fragment
    coexisting_individual_fragments : list
        List of fragment objects that coexist (in frames) with self and that represent an individual
    number_of_coexisting_individual_fragments : int
        Number of individual fragments coexisting with self
    """
    def __init__(self, fragment_identifier = None, start_end = None,
                        blob_hierarchy_in_starting_frame = None, images = None,
                        bounding_box_in_frame_coordinates = None,
                        centroids = None, areas = None,
                        is_an_individual = None, is_a_crossing = None,
                        number_of_animals = None,
                        user_generated_identities = None):
        self.identifier = fragment_identifier
        self.start_end = start_end
        self.blob_hierarchy_in_starting_frame = blob_hierarchy_in_starting_frame
        self.images = images
        self.bounding_box_in_frame_coordinates = bounding_box_in_frame_coordinates
        self.centroids = np.asarray(centroids)
        if centroids is not None:
            self.set_distance_travelled()
        self.areas = np.asarray(areas)
        self.is_an_individual = is_an_individual
        self.is_a_crossing = is_a_crossing
        self.number_of_animals = number_of_animals
        self.possible_identities = range(1, self.number_of_animals + 1)
        self._is_in_a_global_fragment = False
        self._used_for_training = False
        self._used_for_pretraining = False
        self._acceptable_for_training = None
        self._temporary_id = None
        self._identity = None
        self._identity_corrected_solving_jumps = None
        self._user_generated_identities = user_generated_identities
        self._identity_is_fixed = False
        self._accumulated_globally = False
        self._accumulated_partially = False
        self._accumulation_step = None

    def reset(self, roll_back_to = None):
        """Reset attributes of self to a specific part of the algorithm

        Parameters
        ----------
        roll_back_to : str
            Reset all the attributes up to the process specified in input.
            'fragmentation', 'pretraining', 'accumulation', 'assignment'

        """
        if roll_back_to == 'fragmentation' or roll_back_to == 'pretraining':
            self._used_for_training = False
            if roll_back_to == 'fragmentation': self._used_for_pretraining = False
            self._acceptable_for_training = None
            self._temporary_id = None
            self._identity = None
            self._identity_corrected_solving_jumps = None
            self._identity_is_fixed = False
            self._accumulated_globally = False
            self._accumulated_partially = False
            self._accumulation_step = None
            attributes_to_delete = ['_frequencies',
                                    '_P1_vector', '_certainty',
                                    '_is_certain',
                                    '_P1_below_random', '_non_consistent']
            delete_attributes_from_object(self, attributes_to_delete)
        elif roll_back_to == 'accumulation':
            self._identity_is_fixed = False
            attributes_to_delete = []
            if not self.used_for_training:
                self._identity = None
                self._identity_corrected_solving_jumps = None
                attributes_to_delete = ['_frequencies', '_P1_vector']
            attributes_to_delete.extend(['_P2_vector', '_ambiguous_identities',
                                        '_certainty_P2'])
            delete_attributes_from_object(self, attributes_to_delete)
        elif roll_back_to == 'assignment':
            self._user_generated_identity = None
            self._identity_corrected_solving_jumps = None

    @property
    def is_in_a_global_fragment(self):
        return self._is_in_a_global_fragment

    @property
    def used_for_training(self):
        return self._used_for_training

    @property
    def accumulated_globally(self):
        return self._accumulated_globally

    @property
    def accumulated_partially(self):
        return self._accumulated_partially

    @property
    def accumulation_step(self):
        return self._accumulation_step

    @property
    def accumulable(self):
        return self._accumulable

    @property
    def used_for_pretraining(self):
        return self._used_for_pretraining

    @property
    def acceptable_for_training(self):
        return self._acceptable_for_training

    @property
    def frequencies(self):
        return self._frequencies

    @property
    def P1_vector(self):
        return self._P1_vector

    @property
    def P2_vector(self):
        return self._P2_vector

    @property
    def certainty(self):
        return self._certainty

    @property
    def certainty_P2(self):
        return self._certainty_P2

    @property
    def is_certain(self):
        return self._is_certain

    @property
    def temporary_id(self):
        return self._temporary_id

    @property
    def temporary_id_for_pretraining(self):
        return self._temporary_id_for_pretraining

    @property
    def identity(self):
        return self._identity

    @property
    def identity_is_fixed(self):
        return self._identity_is_fixed

    @property
    def identity_corrected_solving_jumps(self):
        return self._identity_corrected_solving_jumps

    @property
    def identities_corrected_closing_gaps(self):
        return self._identities_corrected_closing_gaps

    @property
    def user_generated_identity(self):
        return self._user_generated_identity

    @property
    def final_identities(self):
        if hasattr(self, 'user_generated_identities') and self.user_generated_identities is not None:
            return self.user_generated_identities
        else:
            return self.assigned_identities

    @property
    def assigned_identities(self):
        if hasattr(self, 'identiies_corrected_closing_gaps') and self.identities_corrected_closing_gaps is not None:
            return self.identiies_corrected_closing_gaps
        elif hasattr(self, 'identity_corrected_solving_jumps') and self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        else:
            return [self.identity]

    @property
    def ambiguous_identities(self):
        return self._ambiguous_identities

    @property
    def potentially_randomly_assigned(self):
        return self._potentially_randomly_assigned

    @property
    def non_consistent(self):
        return self._non_consistent

    @property
    def number_of_images(self):
        return len(self.images)

    @property
    def has_enough_accumulated_coexisting_fragments(self):
        return sum([fragment.used_for_training
                    for fragment in self.coexisting_individual_fragments]) >= self.number_of_coexisting_individual_fragments/2

    def get_attribute_of_coexisting_fragments(self, attribute):
        """Retrieve a spevific attribute from the collection of fragments
        coexisting (in frame) with self

        Parameters
        ----------
        attribute : str
            attribute to retrieve

        Returns
        -------
        list
            attribute specified in `attribute` of the fragments coexisting with self

        """
        return [getattr(fragment,attribute) for fragment in self.coexisting_individual_fragments]

    def set_distance_travelled(self):
        """Computes the distance travelled by the individual in the fragment.
        It is based on the position of the centroids in consecutive images. See
        :attr:`~blob.Blob.centroid`

        """
        if self.centroids.shape[0] > 1:
            self.distance_travelled = np.sum(np.sqrt(np.sum(np.diff(self.centroids, axis = 0)**2, axis = 1)))
        else:
            self.distance_travelled = 0.

    def frame_by_frame_velocity(self):
        """Short summary.

        Returns
        -------
        ndarray
            Frame by frame speed of the individual in the fragment

        """
        return np.sqrt(np.sum(np.diff(self.centroids, axis = 0)**2, axis = 1))

    def compute_border_velocity(self, other):
        """Velocity necessary to cover the space between two fragments. Note that
        these velocities are divided by the number of frames that separate self and other
        outside of the function.

        Parameters
        ----------
        other : <Fragment object>
            A second fragment

        Returns
        -------
        float
            Returns the speed at which an individual should travel to be
            present in both self and other Fragment objects

        """
        centroids = np.asarray([self.centroids[0], other.centroids[-1]])
        if not self.start_end[0] > other.start_end[1]:
            centroids = np.asarray([self.centroids[-1],other.centroids[0]])
        return np.sqrt(np.sum(np.diff(centroids, axis = 0)**2, axis = 1))[0]

    def are_overlapping(self, other):
        """Short summary.

        Parameters
        ----------
        other :  <Fragment object>
            A second fragment

        Returns
        -------
        bool
            True if self and other coexist in at least one frame

        """
        (s1,e1), (s2,e2) = self.start_end, other.start_end
        return s1 < e2 and e1 > s2

    def get_coexisting_individual_fragments_indices(self, fragments):
        """Get the list of fragment objects representing and individual (i.e.
        not representing a crossing where two or more animals are touching) and
        coexisting (in frame) with self

        Parameters
        ----------
        fragments : list
            List of all the fragments in the video

        """
        self.coexisting_individual_fragments = [fragment for fragment in fragments
                                            if fragment.is_an_individual and self.are_overlapping(fragment)
                                            and fragment is not self]
        self.number_of_coexisting_individual_fragments = len(self.coexisting_individual_fragments)

    def check_consistency_with_coexistent_individual_fragments(self, temporary_id):
        """Check that the temporary identity assigned to the fragment is
        consistent with respect to the identities already assigned to the
        fragments coexisting (in frame) with it

        Parameters
        ----------
        temporary_id : int
            Temporary identity assigned to the fragment

        Returns
        -------
        bool
            True if the identification of self with `temporary_id` does not
            cause any duplication

        """
        for coexisting_fragment in self.coexisting_individual_fragments:
            if coexisting_fragment.temporary_id == temporary_id:
                return False
        return True

    def compute_identification_statistics(self, predictions, softmax_probs, number_of_animals = None):
        """Computes the statistics necessary to the identification of the
        fragment

        Parameters
        ----------
        predictions : ndarray
            array of shape [number_of_images_in_fragment, 1] whose components
            are the argmax(softmax_probs) per image
        softmax_probs : ndarray
            array of shape [number_of_images_in_fragment, number_of_animals]
            whose rows are the result of applying the softmax function to the
            predictions outputted by the idCNN per image
        number_of_animals : int
            Description of parameter `number_of_animals`.

        See Also
        --------
        :meth:`compute_identification_frequencies_individual_fragment`
        :meth:`compute_P1_from_frequencies`
        :meth:`compute_median_softmax`
        :meth:`compute_certainty_of_individual_fragment`

        """
        assert self.is_an_individual
        number_of_animals = self.number_of_animals if number_of_animals is None else number_of_animals
        self._frequencies = self.compute_identification_frequencies_individual_fragment(predictions, number_of_animals)
        self._P1_vector = self.compute_P1_from_frequencies(self.frequencies)
        median_softmax = self.compute_median_softmax(softmax_probs, number_of_animals)
        self._certainty = self.compute_certainty_of_individual_fragment(self._P1_vector,median_softmax)

    def set_P1_vector_accumulated(self):
        """If self has been used for training its P1_vector is modified to be
        a vector of zeros with a single component set to 1 in the
        :attr:`temporary_id` position
        """
        assert self.used_for_training and self.is_an_individual
        self._P1_vector = np.zeros(len(self.P1_vector))
        self._P1_vector[self.temporary_id] = 1.

    @staticmethod
    def get_possible_identities(P2_vector):
        """Check if P2 has two identical maxima. In that case returns the indices.
        Else return false
        """
        maxima_indices = np.where(P2_vector == np.max(P2_vector))[0]
        return maxima_indices + 1, np.max(P2_vector)

    def assign_identity(self):
        """Assigns the identity to self by considering the fragments coexisting
        with it. If the certainty of the identification is high enough it sets
        the identity of self to be fixed
        """
        assert self.is_an_individual
        if self.used_for_training and not self._identity_is_fixed:
            self._identity_is_fixed = True
        elif not self._identity_is_fixed:
            possible_identities, max_P2 = self.get_possible_identities(self.P2_vector)
            if len(possible_identities) > 1:
                self._identity = 0
                self.zero_identity_assigned_by_P2 = True
                self._ambiguous_identities = possible_identities
            else:
                if max_P2 > conf.FIXED_IDENTITY_THRESHOLD:
                    self._identity_is_fixed = True
                self._identity = possible_identities[0]
                self._P1_vector = np.zeros(len(self.P1_vector))
                self._P1_vector[self.identity - 1] = 1.
                self.recompute_P2_of_coexisting_fragments()

    def recompute_P2_of_coexisting_fragments(self):
        """Updates the P2 of the fragments coexisting with self
        (see :attr:`coexisting_individual_fragments`) if their identity is not
        fixed (see :attr:`identity_is_fixed`)
        """
        # The P2 of fragments with fixed identity won't be recomputed
        # due to the condition in assign_identity() (second line)
        [fragment.compute_P2_vector() for fragment in self.coexisting_individual_fragments]

    def compute_P2_vector(self):
        """Computes the P2_vector related to self. It is based on :attr:`coexisting_individual_fragments`
        """
        coexisting_P1_vectors = np.asarray([fragment.P1_vector for fragment in self.coexisting_individual_fragments])
        numerator = np.asarray(self.P1_vector) * np.prod(1. - coexisting_P1_vectors, axis = 0)
        denominator = np.sum(numerator)
        if denominator != 0:
            self._P2_vector = numerator / denominator
            P2_vector_ordered = np.sort(self.P2_vector)
            P2_first_max = P2_vector_ordered[-1]
            P2_second_max = P2_vector_ordered[-2]
            self._certainty_P2 = conf.MAX_FLOAT if P2_second_max == 0 else P2_first_max / P2_second_max
        else:
            self._P2_vector = np.zeros(self.number_of_animals)
            self._certainty_P2 = 0.

    @staticmethod
    def compute_identification_frequencies_individual_fragment(predictions, number_of_animals):
        """Counts the argmax of predictions per row

        Parameters
        ----------
        predictions : ndarray
            array of shape [n, 1]
        number_of_animals : int
            number of animals to be tracked

        Returns
        -------
        ndarray
            array of shape [1, number_of_animals], whose ith component counts how
            many predictions have maximum components at the index i

        """
        return np.asarray([np.sum(predictions == i)
                            for i in range(1, number_of_animals+1)]) # The predictions come from 1 to number_of_animals + 1

    @staticmethod
    def compute_P1_from_frequencies(frequencies):
        """Given the frequencies of a individual fragment
        computer the P1 vector. P1 is the softmax of the frequencies with base 2
        for each identity.
        """
        P1_of_fragment = 1. / np.sum(2.**(np.tile(frequencies, (len(frequencies),1)).T - np.tile(frequencies, (len(frequencies),1))), axis = 0)
        return P1_of_fragment

    @staticmethod
    def compute_median_softmax(softmax_probs, number_of_animals):
        """Given the softmax of the predictions outputted by the network, it
        computes their median according to the argmax of the softmaxed
        predictions per image

        Parameters
        ----------
        softmax_probs : ndarray
            array of shape [number_of_images_in_fragment, number_of_animals]
            whose rows are the result of applying the softmax function to the
            predictions outputted by the idCNN per image
        number_of_animals : int
            number of animals to be tracked

        Returns
        -------
        type
            Median of argmax(softmax_probs) per image

        """
        softmax_probs = np.asarray(softmax_probs)
        #jumps are fragment composed by a single image, thus:
        if len(softmax_probs.shape) == 1:
            softmax_probs = np.expand_dims(softmax_probs, axis = 1)
        max_softmax_probs = np.max(softmax_probs, axis = 1)
        argmax_softmax_probs = np.argmax(softmax_probs, axis = 1)
        softmax_median = np.zeros(number_of_animals)
        for i in np.unique(argmax_softmax_probs):
            softmax_median[i] = np.median(max_softmax_probs[argmax_softmax_probs == i])
        return softmax_median

    @staticmethod
    def compute_certainty_of_individual_fragment(P1_vector, median_softmax):
        """Computes the certainty given the P1_vector of the fragment by
        using the output of :meth:`compute_median_softmax`

        Parameters
        ----------
        P1_vector : ndarray
            array with shape [1, number_of_animals] computed from frequencies by :meth:`compute_identification_statistics`
        median_softmax : ndarray
            Median of argmax(softmax_probs) per image

        Returns
        -------
        float
            Fragment's certainty

        """
        argsort_p1_vector = np.argsort(P1_vector)
        sorted_p1_vector = P1_vector[argsort_p1_vector]
        sorted_softmax_probs = median_softmax[argsort_p1_vector]
        certainty = np.diff(np.multiply(sorted_p1_vector,sorted_softmax_probs)[-2:])/np.sum(sorted_p1_vector[-2:])
        return certainty[0]

    def get_neighbour_fragment(self, fragments, scope, number_of_frames_in_direction = 0):
        """If it exist, gets the fragment in the list of all fragment whose
        identity is the identity assigned to self and whose starting frame is
        the starting frame of self + 1, or ending frame is the ending frame of
        self - 1

        Parameters
        ----------
        fragments : list
            List of all the fragments in the video
        scope : str
            If "to_the_future" looks for the consecutive fragment wrt to self,
            if "to_the_past" looks for the fragment the precedes self
        number_of_frames_in_direction : int
            Distance (in frame) at which the previous or next fragment has to
            be

        Returns
        -------
        <Fragment object>
            The neighbouring fragment with respect to self in the direction
            specified by scope if it exists. Otherwise None

        """
        if scope == 'to_the_past':
            neighbour = [fragment for fragment in fragments
                            if fragment.is_an_individual and len(fragment.assigned_identities) == 1
                            and fragment.assigned_identities[0] == self.assigned_identities[0]
                            and self.start_end[0] - fragment.start_end[1] == number_of_frames_in_direction]
        elif scope == 'to_the_future':
            neighbour = [fragment for fragment in fragments
                            if fragment.is_an_individual and len(fragment.assigned_identities) == 1
                            and fragment.assigned_identities[0] == self.assigned_identities[0]
                            and fragment.start_end[0] - self.start_end[1] == number_of_frames_in_direction]

        assert len(neighbour) < 2
        return neighbour[0] if len(neighbour) == 1 else None

    def set_partially_or_globally_accumulated(self, accumulation_strategy):
        """Sets :attr:`accumulated_globally` and :attr:`accumulated_partially`
        according to `accumulation_strategy`

        Parameters
        ----------
        accumulation_strategy : str
            Can be "global" or "partial"

        """
        if accumulation_strategy == 'global':
            self._accumulated_globally = True
        elif accumulation_strategy == 'partial':
            self._accumulated_partially = True



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
from tqdm import tqdm

from idtrackerai.list_of_fragments import ListOfFragments

def compute_model_velocity(fragments, number_of_animals, percentile = None):
    """computes the 2 * (percentile) of the distribution of velocities of identified fish.
    params
    -----
    blobs_in_video: list of blob objects
        collection of blobs detected in the video.
    number_of_animals int
    percentile int
    -----
    return
    -----
    float
    2 * np.max(distance_travelled_in_individual_fragments) if percentile is None
    2 * percentile(velocity distribution of identified animals) otherwise
    """
    distance_travelled_in_individual_fragments = []

    for fragment in tqdm(fragments, desc = "computing velocity model"):
        if fragment.is_an_individual:
            distance_travelled_in_individual_fragments.extend(fragment.frame_by_frame_velocity())

    return 2 * np.max(distance_travelled_in_individual_fragments) if percentile is None else 2 * np.percentile(distance_travelled_in_individual_fragments, percentile)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import gc

import numpy as np
import multiprocessing
import cv2
import h5py
from joblib import Parallel, delayed
from tqdm import tqdm
from scipy import ndimage
from confapp import conf

from idtrackerai.blob import Blob
from idtrackerai.utils.py_utils import flatten, set_mkl_to_single_thread, set_mkl_to_multi_thread
from idtrackerai.utils.segmentation_utils import segment_frame, blob_extractor

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.segmentation")

"""
The segmentation module
"""

def get_videoCapture(video, path, episode_start_end_frames):
    """Gives the VideoCapture (OpenCV) object to read the frames for the segmentation
    and the number of frames to read. If `episode_start_end_frames` is None then a `path` must be
    given as the video is assumed to be splitted in different files (episodes). If the `path`
    is None the video is assumed to be in a single file and the path is read from
    `video`, then `episode_start_end_frames` must be give.

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    path : string
        Path to the video file from where to get the VideoCapture (OpenCV) object
    episode_start_end_frames : tuple
        Tuple (starting_frame, ending_frame) indicanting the start and end of the episode
        when the video is given in a single file

    Returns
    -------
    cap : <VideoCapture object>
        OpenCV object used to read the frames of the video
    number_of_frames_in_episode : int
        Number of frames in the episode of video being segmented
    """
    if path is not None:
        cap = cv2.VideoCapture(path)
        number_of_frames_in_episode = int(cap.get(7))
    elif path is None:
        cap = cv2.VideoCapture(video.video_path)
        number_of_frames_in_episode = episode_start_end_frames[1] - episode_start_end_frames[0] + 1
        cap.set(1,episode_start_end_frames[0])

    return cap, number_of_frames_in_episode


def get_blobs_in_frame(cap, video, segmentation_thresholds,
                       max_number_of_blobs, frame_number, global_frame_number,
                       frame_number_in_video_path,
                       bounding_box_images_path, episode, video_path, pixels_path,
                       save_pixels, save_segmentation_image):
    """Segments a frame read from `cap` according to the preprocessing parameters
    in `video`. Returns a list `blobs_in_frame` with the Blob objects in the frame
    and the `max_number_of_blobs` found in the video so far. Frames are segmented
    in gray scale.

    Parameters
    ----------
    cap : <VideoCapture object>
        OpenCV object used to read the frames of the video
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    segmentation_thresholds : dict
        Dictionary with the thresholds used for the segmentation: `min_threshold`,
        `max_threshold`, `min_area`, `max_area`
    max_number_of_blobs : int
        Maximum number of blobs found in the whole video so far in the segmentation process
    frame_number : int
        Number of the frame being segmented. It is used to print in the terminal the frames
        where the segmentation fails. This frame is the frame of the episode if the video
        is chuncked.
    global_frame_number : int
        This is the frame number in the whole video. It will be different to the frame_number
        if the video is chuncked.


    Returns
    -------
    blobs_in_frame : list
        List of <Blob object> segmented in the current frame
    max_number_of_blobs : int
        Maximum number of blobs found in the whole video so far in the segmentation process

    See Also
    --------
    Video
    Blob
    segment_frame
    blob_extractor
    """
    blobs_in_frame = []
    ret, frame = cap.read()
    if conf.SIGMA_GAUSSIAN_BLURRING is not None:
        frame = cv2.GaussianBlur(frame, (0, 0), conf.SIGMA_GAUSSIAN_BLURRING)

    try:
        if video.resolution_reduction != 1 and ret:
            frame = cv2.resize(frame, None,
                               fx=video.resolution_reduction,
                               fy=video.resolution_reduction,
                               interpolation=cv2.INTER_AREA)
        frameGray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        avIntensity = np.float32(np.mean(np.ma.array(frameGray,
                                                     mask=video.ROI==0)))
        segmentedFrame = segment_frame(frameGray / avIntensity,
                                       segmentation_thresholds['min_threshold'],
                                       segmentation_thresholds['max_threshold'],
                                       video.bkg, video.ROI, video.subtract_bkg)
        # Fill holes in the segmented frame to avoid duplication of contours
        segmentedFrame = ndimage.binary_fill_holes(segmentedFrame).astype('uint8')
        # Find contours in the segmented image
        bounding_boxes, miniframes, centroids, \
            areas, pixels, contours, estimated_body_lengths = \
            blob_extractor(segmentedFrame, frameGray,
                           segmentation_thresholds['min_area'],
                           segmentation_thresholds['max_area'],
                           save_pixels, save_segmentation_image)
    except Exception as e:
        print(e)
        logger.info("An error occurred while reading frame number : %i" %frame_number)
        logger.info("ret: %s" %str(ret))
        logger.info("frame: %s" %str(frame))
        bounding_boxes = []
        miniframes = []
        centroids = []
        areas = []
        pixels = []
        contours = []

    # create blob objects
    for i, bounding_box in enumerate(bounding_boxes):
        if save_segmentation_image == 'DISK':
            with h5py.File(bounding_box_images_path, 'a') as f1:
                f1.create_dataset(str(global_frame_number) + '-' + str(i),
                                  data=miniframes[i])
            miniframes[i] = None
        if save_pixels == 'DISK':
            with h5py.File(pixels_path, 'a') as f2:
                f2.create_dataset(str(global_frame_number) + '-' + str(i),
                                  data=pixels[i])
            pixels[i] = None

        blob = Blob(centroids[i],
                    contours[i],
                    areas[i],
                    bounding_box,
                    bounding_box_image=miniframes[i],
                    bounding_box_images_path=bounding_box_images_path,
                    estimated_body_length=estimated_body_lengths[i],
                    number_of_animals=video.number_of_animals,
                    frame_number=global_frame_number,
                    pixels=pixels[i],
                    pixels_path=pixels_path,
                    in_frame_index=i,
                    video_height=video.height,
                    video_width=video.width,
                    video_path=video_path,
                    frame_number_in_video_path=frame_number_in_video_path,
                    resolution_reduction=video.resolution_reduction)
        blobs_in_frame.append(blob)

    if len(centroids) > max_number_of_blobs:
        max_number_of_blobs = len(centroids)

    return blobs_in_frame, max_number_of_blobs


def frame_in_intervals(frame_number, intervals):
    """Returns True if a frame is inside of one of the frame intervals. Otherwise
    returns False

    Parameters
    ----------
    frame_number : int
        Number of the frame to be checked.
    intervals : list
        List of intervals where to check for the frame
    """
    for interval in intervals:
        if frame_number >= interval[0] and frame_number <= interval[1]:
            return True
    return False


def segment_episode(video, segmentation_thresholds,
                    path=None, episode_start_end_frames=None,
                    save_pixels=None,
                    save_segmentation_image=None):
    """Gets list of blobs segmented in every frame of the episode of the video
    given by `path` (if the video is splitted in different files) or by
    `episode_start_end_frames` (if the video is given in a single file)

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    segmentation_thresholds : dict
        Dictionary with the thresholds used for the segmentation: `min_threshold`,
        `max_threshold`, `min_area`, `max_area`
    path : string
        Path to the video file from where to get the VideoCapture (OpenCV) object
    episode_start_end_frames : tuple
        Tuple (starting_frame, ending_frame) indicanting the start and end of the episode
        when the video is given in a single file

    Returns
    -------
    blobs_in_episode : list
        List of `blobs_in_frame` of the episode of the video being segmented
    max_number_of_blobs : int
        Maximum number of blobs found in the episode of the video being segmented

    See Also
    --------
    Video
    Blob
    get_videoCapture
    segment_frame
    blob_extractor
    """
    episode = video.in_which_episode(episode_start_end_frames[0])
    if video.paths_to_video_segments is None:
        video_path = video.video_path
    else:
        video_path = video.paths_to_video_segments[episode]
    bounding_box_images_path = None
    if save_segmentation_image == 'DISK':
        bounding_box_images_path = os.path.join(video._segmentation_data_folder,
                                                'episode_images_{}.hdf5'.format(str(episode)))
        if os.path.isfile(bounding_box_images_path):
            os.remove(bounding_box_images_path)

    pixels_path = None
    if save_pixels == 'DISK':
        pixels_path = os.path.join(video._segmentation_data_folder,
                                   'episode_pixels_{}.hdf5'.format(str(episode)))
        if os.path.isfile(pixels_path):
            os.remove(pixels_path)

    blobs_in_episode = []
    cap, number_of_frames_in_episode = get_videoCapture(video, path, episode_start_end_frames)
    max_number_of_blobs = 0
    frame_number = 0
    while frame_number < number_of_frames_in_episode:

        global_frame_number = episode_start_end_frames[0] + frame_number
        if video.paths_to_video_segments is None:
            frame_number_in_video_path = global_frame_number
        else:
            frame_number_in_video_path = frame_number
        if video.tracking_interval is None or frame_in_intervals(global_frame_number, video.tracking_interval):
            blobs_in_frame, max_number_of_blobs = \
                get_blobs_in_frame(cap, video, segmentation_thresholds,
                                   max_number_of_blobs, frame_number,
                                   global_frame_number,
                                   frame_number_in_video_path,
                                   bounding_box_images_path,
                                   episode, video_path, pixels_path,
                                   save_pixels, save_segmentation_image)
        else:
            ret, _ = cap.read()
            blobs_in_frame = []
        #store all the blobs encountered in the episode
        blobs_in_episode.append(blobs_in_frame)
        frame_number += 1
    cap.release()
    #cv2.destroyAllWindows()
    gc.collect()
    return blobs_in_episode, max_number_of_blobs


def segment(video):
    """Segment the video giving a list of `blobs` for every frame in the video.
    If a video is given as a set of files (episodes), those files are used to
    parallelise the segmentation process. If a video is given in a single file
    the list of indices `video.episodes_start_end` is used for the parallelisation

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading

    Returns
    -------
    blobs_in_video : list
        List of `blobs_in_frame` for all the frames of the video

    See Also
    --------
    segment_episode
    """
    # avoid computing with all the cores in very large videos. It fills the RAM.
    num_cpus = int(multiprocessing.cpu_count())
    num_jobs = conf.NUMBER_OF_JOBS_FOR_SEGMENTATION
    # num_jobs = 1
    if conf.NUMBER_OF_JOBS_FOR_SEGMENTATION is None:
        num_jobs = 1
    elif conf.NUMBER_OF_JOBS_FOR_SEGMENTATION < 0:
        num_jobs = (num_cpus + 1 + num_jobs)

    #init variables to store data
    blobs_in_video = []
    maximum_number_of_blobs_in_episode = []
    segmentation_thresholds = {'min_threshold': video.min_threshold,
                                'max_threshold': video.max_threshold,
                                'min_area': video.min_area,
                                'max_area': video.max_area}

    set_mkl_to_single_thread()
    if not video.paths_to_video_segments:
        logger.info('There is only one path, segmenting by frame indices')
        #Spliting episodes_start_end in sublists for parallel processing
        episodes_start_end_sublists = [video.episodes_start_end[i:i+num_jobs]
                                        for i in range(0,len(video.episodes_start_end),num_jobs)]

        for episodes_start_end_sublist in tqdm(episodes_start_end_sublists, desc='Segmentation progress'):
            OupPutParallel = Parallel(n_jobs=conf.NUMBER_OF_JOBS_FOR_SEGMENTATION)(
                                delayed(segment_episode)(video, segmentation_thresholds, None, episode_start_end_frames,
                                                         conf.SAVE_PIXELS, conf.SAVE_SEGMENTATION_IMAGE)
                                for episode_start_end_frames in episodes_start_end_sublist)
            blobs_in_episode = [out[0] for out in OupPutParallel]
            maximum_number_of_blobs_in_episode.append([out[1] for out in OupPutParallel])
            blobs_in_video.append(blobs_in_episode)
    else:
        #splitting videoPaths list into sublists
        pathsSubLists = [video.paths_to_video_segments[i: i + num_jobs]
                         for i in range(0,len(video.paths_to_video_segments), num_jobs)]
        episodes_start_end_sublists = [video.episodes_start_end[i:i+num_jobs]
                                       for i in range(0,len(video.episodes_start_end), num_jobs)]

        for pathsSubList, episodes_start_end_sublist in tqdm(list(zip(pathsSubLists, episodes_start_end_sublists)), desc='Segmentation progress'):
            OupPutParallel = Parallel(n_jobs=conf.NUMBER_OF_JOBS_FOR_SEGMENTATION)(
                                delayed(segment_episode)(video, segmentation_thresholds, path, episode_start_end_frames,
                                                         conf.SAVE_PIXELS, conf.SAVE_SEGMENTATION_IMAGE)
                                for path, episode_start_end_frames in zip(pathsSubList, episodes_start_end_sublist))
            blobs_in_episode = [out[0] for out in OupPutParallel]
            maximum_number_of_blobs_in_episode.append([out[1] for out in OupPutParallel])
            blobs_in_video.append(blobs_in_episode)
    set_mkl_to_multi_thread()
    video._maximum_number_of_blobs = max(flatten(maximum_number_of_blobs_in_episode))
    #blobs_in_video is flattened to obtain a list of blobs per episode and then the list of all blobs
    blobs_in_video = flatten(flatten(blobs_in_video))
    return blobs_in_video


def resegment(video, frame_number, list_of_blobs,
              new_segmentation_thresholds):
    """Updates the `list_of_blobs` for a particular `frame_number` by performing
    a segmentation with `new_segmentation_thresholds`. This function is called for
    the frames in which the number of blobs is higher than the number of animals
    stated by the user.

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    frame_number : int
        Number of the frame to update with the new segmentation
    list_of_blobs : <ListOfBlobs object>
        Object containing the list of blobs segmented in the video in each frame
    new_segmentation_thresholds : dict
        Dictionary with the thresholds used for the new segmentation: `min_threshold`,
        `max_threshold`, `min_area`, `max_area`

    Returns
    -------
    number_of_blobs_in_frame : int
        Number of blobs found in the frame

    See Also
    --------
    get_videoCapture
    get_blobs_in_frame
    """
    episode_number = video.in_which_episode(frame_number)
    if video.paths_to_video_segments is None:
        video_path = video.video_path
    else:
        video_path = video.paths_to_video_segments[episode]
    if not video.paths_to_video_segments:
        cap, _ = get_videoCapture(video, None, video.episodes_start_end[episode_number])
        cap.set(1, frame_number)
        frame_number_in_video_path = frame_number
    else:
        path = video.paths_to_video_segments[episode_number]
        cap, _ = get_videoCapture(video, path, None)
        start = video.episodes_start_end[episode_number][0]
        cap.set(1,frame_number - start)
        frame_number_in_video_path = frame_number-start

    episode = video.in_which_episode(frame_number)
    bounding_box_images_path = None
    if conf.SAVE_SEGMENTATION_IMAGE == 'DISK':
        bounding_box_images_path = os.path.join(video._segmentation_data_folder,
                                    's_images_{}.hdf5'.format(str(episode)))
        with h5py.File(bounding_box_images_path, 'a') as f:
            images_to_delete = [d_name for d_name in f.keys()
                                if d_name.split('-')[0] == str(frame_number)]
            for im in images_to_delete:
                del f[im]

    pixels_path = None
    if conf.SAVE_PIXELS == 'DISK':
        pixels_path = os.path.join(video._segmentation_data_folder,
                                   'episode_pixels_{}.hdf5'.format(str(episode)))
        with h5py.File(pixels_path, 'a') as f:
            pixels_to_delete = [d_name for d_name in f.keys()
                                if d_name.split('-')[0] == str(frame_number)]
            for px in pixels_to_delete:
                del f[px]


    blobs_in_resegmanted_frame, \
        number_of_blobs_in_frame = \
        get_blobs_in_frame(cap, video, new_segmentation_thresholds,
                           0, frame_number, frame_number,
                           frame_number_in_video_path,
                           bounding_box_images_path, episode,
                           video_path, pixels_path,
                           conf.SAVE_PIXELS, conf.SAVE_SEGMENTATION_IMAGE)
    list_of_blobs.blobs_in_video[frame_number] = blobs_in_resegmanted_frame
    cap.release()
    cv2.destroyAllWindows()
    return number_of_blobs_in_frame



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import psutil

import numpy as np
from confapp import conf

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.get_predictions_crossings")

class GetPredictionCrossigns(object):
    def __init__(self, net):
        # Data set
        self.net = net
        self._softmax_probs = []
        self._predictions = []
        self._fc_vectors = []
        self.batch_size = conf.BATCH_SIZE_PREDICTIONS_DCD


    def next_batch_test(self, batch_size):
        """Return the next `batch_size` examples from this data set."""
        start = self.number_of_image_predicted
        self.number_of_image_predicted += batch_size
        end = self.number_of_image_predicted
        return self.test_images[start:end]

    def predict(self, test_images):
        self.test_images = test_images
        self.number_of_image_predicted = 0
        predictions = []
        while self.number_of_image_predicted < len(test_images):
            predictions.extend(self.net.prediction(self.next_batch_test(batch_size = conf.BATCH_SIZE_PREDICTIONS_DCD)))

        return predictions

    def get_all_predictions(self, test_set):
        # compute maximum number of images given the available RAM
        image_size_bytes = np.prod(test_set.image_size)*4 #XXX:Check it out!!!! np.prod(test_set.image_size**2)*4
        number_of_images_to_be_fitted_in_RAM = len(test_set.test)
        num_images_that_can_fit_in_RAM = 100000#int(psutil.virtual_memory().available*.2/image_size_bytes)
        if number_of_images_to_be_fitted_in_RAM > num_images_that_can_fit_in_RAM:
            logger.debug("There is NOT enough RAM to host %i images" %number_of_images_to_be_fitted_in_RAM)
            number_of_predictions_retrieved = 0
            predictions = []
            i = 0
            while number_of_predictions_retrieved < number_of_images_to_be_fitted_in_RAM:
                images = test_set.generate_test_images(interval = (i*num_images_that_can_fit_in_RAM, (i+1)*num_images_that_can_fit_in_RAM))
                predictions.extend(self.predict(images))
                number_of_predictions_retrieved = len(predictions)
                i += 1
        else:
            logger.debug("There is enough RAM to host %i images" %number_of_images_to_be_fitted_in_RAM)
            logger.info("getting predictions...")
            test_images = test_set.generate_test_images()
            predictions = self.predict(test_images)
        return predictions

    # def get_predictions_from_images(self, images):
    #     image_size_bytes = np.prod(images.shape[1:])*4
    #     number_of_images_to_be_fitted_in_RAM = len(images)
    #     num_images_that_can_fit_in_RAM = int(psutil.virtual_memory().available*.9/image_size_bytes)
    #     if number_of_images_to_be_fitted_in_RAM > num_images_that_can_fit_in_RAM:
    #         logger.debug("There is NOT enough RAM to host %i images" %number_of_images_to_be_fitted_in_RAM)
    #         number_of_predictions_retrieved = 0
    #         predictions = []
    #         i = 0
    #         while number_of_predictions_retrieved < number_of_images_to_be_fitted_in_RAM:
    #             images_batch = images[i*num_images_that_can_fit_in_RAM : (i+1)*num_images_that_can_fit_in_RAM]
    #             predictions.extend(self.predict(images_batch))
    #             number_of_predictions_retrieved = len(predictions)
    #             i += 1
    #     else:
    #         logger.debug("There is enough RAM to host %i images" %number_of_images_to_be_fitted_in_RAM)
    #         logger.info("getting predictions...")
    #         predictions = self.predict(images)
    #     return predictions



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
from confapp import conf

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.stop_training_criteria_crossings")

class Stop_Training(object):
    """Stops the training of the network according to the conditions specified
    in __call__
    """
    def __init__(self, epochs_before_checking_stopping_conditions = 10, check_for_loss_plateau = True):
        self.num_epochs = conf.MAXIMUM_NUMBER_OF_EPOCHS_DCD #maximal num of epochs
        self.number_of_classes = 2
        self.epochs_before_checking_stopping_conditions = epochs_before_checking_stopping_conditions
        self.overfitting_counter = 0 #number of epochs in which the network is overfitting before stopping the training
        self.check_for_loss_plateau = check_for_loss_plateau #bool: if true the training is stopped if the loss is not decreasing enough

    def __call__(self, loss_accuracy_training, loss_accuracy_validation, epochs_completed):
        #check that the model did not diverged (nan loss).
        if epochs_completed > 0 and (np.isnan(loss_accuracy_training.loss[-1]) or np.isnan(loss_accuracy_validation.loss[-1])):
            logger.warn('The model diverged with loss NaN, falling back to detecting crossings with the model area')
            return True
        #check if it did not reached the epochs limit
        if epochs_completed > self.num_epochs-1:
            logger.warn('The number of epochs completed is larger than the number of epochs set for training, we stop the training')
            return True
        #check that the model is not overfitting or if it reached a stable saddle (minimum)
        if epochs_completed > self.epochs_before_checking_stopping_conditions:
            current_loss = loss_accuracy_validation.loss[-1]
            previous_loss = np.nanmean(loss_accuracy_validation.loss[-self.epochs_before_checking_stopping_conditions:-1])
            #The validation loss in the first 10 epochs could have exploded but being decreasing.
            if np.isnan(previous_loss): previous_loss = conf.MAX_FLOAT
            losses_difference = (previous_loss-current_loss)
            #check overfitting
            if losses_difference < 0.:
                self.overfitting_counter += 1
                if self.overfitting_counter >= conf.OVERFITTING_COUNTER_THRESHOLD_DCD:
                    logger.info('Overfitting')
                    return True
            else:
                self.overfitting_counter = 0
            #check if the error is not decreasing much
            if self.check_for_loss_plateau:
                if np.abs(losses_difference) < conf.LEARNING_PERCENTAGE_DIFFERENCE_2_DCD * 10**(int(np.log10(current_loss))-1):
                    logger.info('The losses difference is very small, we stop the training\n')
                    return True
            # if the individual accuracies in validation are 1. for all the animals
            if list(loss_accuracy_validation.individual_accuracy[-1]) == list(np.ones(self.number_of_classes)):
                logger.info('The individual accuracies in validation is 1. for all the classes, we stop the training\n')
                return True
            # if the validation loss is 0.
            if previous_loss == 0. or current_loss == 0.:
                logger.info('The validation loss is 0., we stop the training')
                return True

        return False



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import numpy as np
from pprint import pprint
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.blob import Blob
from idtrackerai.groundtruth_utils.generate_individual_groundtruth import IndividualGroundTruth, GroundTruthBlob
if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.compute_statistics_against_groundtruth")


def compare_tracked_individual_against_groundtruth(blobs_in_individual_groundtruth,
                                                    individual_blobs, individual_groundtruth_id):
    comparison_keys = ['accuracy', 'frames_with_errors', 'mistaken_identities']
    comparison_info = {key: [] for key in comparison_keys}

    for blob_gt, blob in zip(blobs_in_individual_groundtruth, individual_blobs):
        if blob_gt.identity != blob.assigned_identity:
            comparison_info['frames_with_errors'].append(blob.frame_number)
            comparison_info['mistaken_identities'].append(blob.assigned_identity)

    comparison_info['accuracy'] = 1 -len(comparison_info['mistaken_identities']) / len(blobs_in_individual_groundtruth)
    number_of_assigned_blobs = len([blob for blob in blobs_in_individual_groundtruth if blob.assigned_identity != 0])
    number_of_mistaken_identified_blobs = len([identity for identity in comparison_info['mistaken_identities'] if identity != 0])
    comparison_info['accuracy_assigned'] = 1 - number_of_mistaken_identified_blobs / number_of_assigned_blobs
    comparison_info['id'] = individual_groundtruth_id
    return comparison_info

def check_groundtruth_consistency(blobs_in_individual_groundtruth,
                                    individual_groundtruth_id, individual_blobs,
                                    individual_id):
    non_matching_error = "The length of the collections of the ground truth individual and the selected one do not match"
    if individual_groundtruth_id != individual_id or\
        len(blobs_in_individual_groundtruth) != len(individual_blobs):
        raise ValueError(non_matching_error)

def get_individual_accuracy_wrt_groundtruth(video, blobs_in_individual_groundtruth, individual_blobs = None):
    individual_groundtruth_id = blobs_in_individual_groundtruth[0].identity
    individual_id = individual_groundtruth_id
    if individual_blobs is None:
        individual_blobs = blobs_in_individual_groundtruth
    else:
        check_groundtruth_consistency(blobs_in_individual_groundtruth,
                                    individual_groundtruth_id,
                                    individual_blobs,
                                    individual_id)
    return compare_tracked_individual_against_groundtruth(blobs_in_individual_groundtruth,
                                                individual_blobs, individual_groundtruth_id)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

from confapp import conf

class ModelArea(object):
    """Model of the area used to perform a first discrimination between blobs
    representing single individual and multiple touching animals (crossings)

    Attributes
    ----------

    median : float
        median of the area of the blobs segmented from portions of the video in
        which all the animals are visible (not touching)
    mean : float
        mean of the area of the blobs segmented from portions of the video in
        which all the animals are visible (not touching)
    std : float
        standard deviation of the area of the blobs segmented from portions of
        the video in which all the animals are visible (not touching)
    std_tolerance : int
        tolerance factor
    """
    def __init__(self, mean, median, std):
        self.median = median
        self.mean = mean
        self.std = std
        self.std_tolerance = conf.STD_TOLERANCE

    def __call__(self, area, std_tolerance = None):
        if std_tolerance is not None:
            self.std_tolerance = std_tolerance
        return (area - self.median) < self.std_tolerance * self.std



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.network_params_crossings")

class NetworkParams_crossings(object):
    def __init__(self,number_of_classes,
                architecture = None,
                learning_rate = None, keep_prob = None,
                use_adam_optimiser = False, scopes_layers_to_optimize = None,
                restore_folder = None, save_folder = None, knowledge_transfer_folder = None,
                image_size = None):

        self.number_of_classes = number_of_classes
        self.learning_rate = learning_rate
        self.architecture = architecture
        self.keep_prob = keep_prob
        self._restore_folder = restore_folder
        self._save_folder = save_folder
        self._knowledge_transfer_folder = knowledge_transfer_folder
        self.use_adam_optimiser = use_adam_optimiser
        self.image_size = image_size

    @property
    def restore_folder(self):
        return self._restore_folder

    @restore_folder.setter
    def restore_folder(self, path):
        assert os.path.isdir(path)
        self._restore_folder = path

    @property
    def save_folder(self):
        return self._save_folder

    @save_folder.setter
    def save_folder(self, path):
        if not os.path.isdir(path):
            os.path.makedirs(path)
        self._save_folder = path

    @property
    def knowledge_transfer_folder(self):
        return self._knowledge_transfer_folder

    @knowledge_transfer_folder.setter
    def knowledge_transfer_folder(self, path):
        assert os.path.isdir(path)
        self._knowledge_transfer_folder = path



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
from idtrackerai.utils.py_utils import  *
import tensorflow as tf
import numpy as np
import re

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.cnn_utils")

# ****************************************************************************
# Tensorboard
# *****************************************************************************

# def _activation_summary(x):
#   """Taken from [1]_
#   Helper to create summaries for activations.
#   Creates a summary that provides a histogram of activations.
#   Creates a summary that measure the sparsity of activations.
#   .. [1] https://www.programcreek.com/python/example/90327/tensorflow.scalar_summary
#   """
#   # Remove 'tower_[0-9]/' from the name in case this is a multi-GPU training
#   # session. This helps the clarity of presentation on tensorboard.
#   tensor_name = re.sub('%s_[0-9]*/' % 'tower', '', x.op.name)
#   tf.summary.histogram(tensor_name + '/activations', x)
#   tf.summary.scalar(tensor_name + '/sparsity', tf.nn.zero_fraction(x))
#
# def put_kernels_on_grid(kernel, grid, pad=1):
#     '''Taken from [2]_
#     Visualize conv. features as an image (mostly for the 1st layer).
#     Place kernel into a grid, with some paddings between adjacent filters.
#     Args:
#       kernel:            tensor of shape [Y, X, NumChannels, NumKernels]
#       (grid_Y, grid_X):  shape of the grid. Require: NumKernels == grid_Y * grid_X
#                            User is responsible of how to break into two multiples.
#       pad:               number of black pixels around each filter (between them)
#
#     Return:
#       Tensor of shape [(Y+pad)*grid_Y, (X+pad)*grid_X, NumChannels, 1].
#
#     .. [2] https://gist.github.com/kukuruza/03731dc494603ceab0c5
#     '''
#     grid_Y, grid_X = grid
#     # pad X and Y
#     x1 = tf.pad(kernel, tf.constant( [[pad,pad],[pad,pad],[0,0],[0,0]] ))
#     # X and Y dimensions, w.r.t. padding
#     Y = kernel.get_shape()[0] + 2*pad
#     X = kernel.get_shape()[1] + 2*pad
#     NumChannels = kernel.get_shape()[2]
#     # put NumKernels to the 1st dimension
#     x2 = tf.transpose(x1, (3, 0, 1, 2))
#     # organize grid on Y axis
#     x3 = tf.reshape(x2, tf.stack([grid_X, Y * grid_Y, X, NumChannels]))
#     # switch X and Y axes
#     x4 = tf.transpose(x3, (0, 2, 1, 3))
#     # organize grid on X axis
#     x5 = tf.reshape(x4, tf.stack([1, X * grid_X, Y * grid_Y, NumChannels]))
#     # back to normal order (not combining with the next step for clarity)
#     x6 = tf.transpose(x5, (2, 1, 3, 0))
#     # to tf.summary.image order [batch_size, height, width, channels],
#     #   where in this case batch_size == 1
#     x7 = tf.transpose(x6, (3, 0, 1, 2))
#     # scale to [0, 1]
#     x_min = tf.reduce_min(x7)
#     x_max = tf.reduce_max(x7)
#     x8 = (x7 - x_min) / (x_max - x_min)
#
#     return x8

# ****************************************************************************
# CNN wrappers
# *****************************************************************************

def computeVolume(width, height, strides):
    c1 = float(strides[1])
    c2 = float(strides[2])
    widthS = int(np.ceil(width/c1))
    heightS = int(np.ceil(height/c2))
    return widthS, heightS

def buildConv2D(scopeName, inputWidth, inputHeight, inputDepth, inputConv ,filter_size, n_filters, stride, pad):
    w,h = computeVolume(inputWidth, inputHeight, stride)
    with tf.variable_scope(scopeName) as scope:
        W = tf.get_variable(
            'weights',
            [filter_size, filter_size, inputDepth, n_filters],
            # initializer=tf.random_normal_initializer(mean=0.0,stddev=0.1)
            initializer=tf.contrib.layers.xavier_initializer_conv2d(seed=0.)
            )
        b = tf.get_variable(
            'biases',
            [n_filters],
            # initializer=tf.random_normal_initializer(mean=0.0,stddev=0.1)
            initializer=tf.constant_initializer(0.0)
            )
        conv = tf.nn.conv2d(
                    input=inputConv,
                     filter=W,
                     strides=stride,
                     padding=pad)
        convb = tf.nn.bias_add(conv, b, name = scope.name)

        # _activation_summary(convb)
        # 
        # with tf.variable_scope('visualization') :
        #     grid_x = int(np.sqrt(n_filters))
        #     grid_y = grid_x  # to get a square grid for 64 conv1 features
        #     WtoPlot = tf.slice(W, [0, 0, 0, 0], [filter_size, filter_size, 1, n_filters])
        #     grid = put_kernels_on_grid(WtoPlot, (grid_y, grid_x))
        #     tf.summary.image(scopeName + '/features', grid, max_outputs=1)
        #     convbToPlot = tf.slice(convb, [0, 0, 0, 0], [-1, w, h, 1])
        #     # this will display random images
        #     tf.summary.image(scopeName + '/output', convbToPlot, max_outputs=10)
    return convb, w, h, W

def maxpool2d(name,inputWidth, inputHeight, inputPool, pool=2 ,
                stride=[1,2,2,1] ,pad='VALID'):
    max_pool = tf.nn.max_pool(inputPool,
        ksize=[1, pool, pool, 1],
        strides=stride,
        padding=pad,
        name = name
        )
    w, h = computeVolume(inputWidth, inputHeight, stride)
    return max_pool, w, h

def buildFc(scopeName, inputFc, height, width, n_filters, n_fc, keep_prob):
    with tf.variable_scope(scopeName) as scope:
        W = tf.get_variable(
            'weights',
            [height * width * n_filters, n_fc],
            initializer=tf.contrib.layers.xavier_initializer(seed=0.)
            )
        b = tf.get_variable(
            'biases',
            [n_fc],
            initializer=tf.constant_initializer(0.0)
            )
        fc = tf.add(tf.matmul(inputFc, W), b)
        fc_drop = tf.nn.dropout(fc, keep_prob, name = scope.name)
        # _activation_summary(fc_drop)
    return fc_drop, W

def reLU(scopeName, inputRelu):
    with tf.variable_scope(scopeName) as scope:
        relu = tf.nn.relu(inputRelu, name = scope.name)
        # _activation_summary(relu)
    return relu

def buildSoftMax(scopeName, inputSoftMax, n_fc, classes):
    with tf.variable_scope(scopeName) as scope:
        W = tf.get_variable(
            'weights',
            [n_fc, classes],
            initializer=tf.contrib.layers.xavier_initializer(seed=0.)
            )
        b = tf.get_variable(
            'biases',
            [classes],
            initializer=tf.constant_initializer(0.0)
            )
        logits = tf.add(tf.matmul(inputSoftMax, W), b, name = scope.name)
        # _activation_summary(logits)
    return logits, W



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logging.getLogger("__main__.store_accuracy_and_loss_crossings")

class Store_Accuracy_and_Loss(object):
    """Store the loss, accuracy and individual accuracy values computed during
    training and validation
    """
    def __init__(self, network, name):
        self._path_to_accuracy_error_data = network.params.save_folder
        self.name = name
        self.loss = []
        self.accuracy = []
        self.individual_accuracy = []
        # if network.is_restoring or network.is_knowledge_transfer:
        #     self.load()

    def append_data(self, loss_value, accuracy_value, individual_accuracy_value):
        self.loss.append(loss_value)
        self.accuracy.append(accuracy_value)
        self.individual_accuracy.append(individual_accuracy_value)

    def plot(self, axes_handles = None, index = 0, color = 'r', plot_now = True, legend_font_color = None, from_GUI = False):
        if from_GUI:
            ax1 = axes_handles[0]
            ax2 = axes_handles[1]
            ax1.plot(range(index,len(self.loss)),self.accuracy[index:], color,label=self.name)
            width = 0.35
            numAnimals = len(self.individual_accuracy[-1])
            if self.name == 'training':
                ind = np.arange(numAnimals) + 1 - width
            elif self.name == 'validation':
                ind = np.arange(numAnimals) + 1

            ax2.bar(ind, self.individual_accuracy[-1], width, color=color, alpha=0.4,label=self.name)
            ax2.set_xlabel('image type (individual or crossings)')
            ax2.set_ylabel('Per class \naccuracy')
            if index == 0:
                legend = ax1.legend()
                if legend_font_color is not None:
                    for text in legend.get_texts():
                        text.set_color(legend_font_color)
                ax1.set_ylabel('Accuracy')
        else:
            ax1 = axes_handles[0]
            ax2 = axes_handles[1]
            ax3 = axes_handles[2]
            ax1.plot(range(index,len(self.loss)), self.loss[index:], color,label=self.name)
            ax2.plot(range(index,len(self.loss)),self.accuracy[index:], color,label=self.name)
            width = 0.35
            numAnimals = len(self.individual_accuracy[-1])
            if self.name == 'training':
                ind = np.arange(numAnimals) + 1 - width
            elif self.name == 'validation':
                ind = np.arange(numAnimals) + 1

            ax3.bar(ind, self.individual_accuracy[-1], width, color=color, alpha=0.4,label=self.name)
            ax3.set_xlabel('image type (individual or crossings)')
            ax3.set_ylabel('Per class \naccuracy')
            if index == 0:
                legend = ax1.legend()
                if legend_font_color is not None:
                    for text in legend.get_texts():
                        text.set_color(legend_font_color)
                ax1.set_ylabel('Loss')
                ax2.set_ylabel('Accuracy')
                ax2.set_xlabel('Epochs')
            if plot_now:
                plt.draw()
                plt.pause(1e-8)

    def save(self):
        np.save(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy'), self.__dict__)

    def load(self):
        if os.path.isfile(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy')):
            loss_accuracy_dictionary = np.load(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy'), allow_pickle=True).item()
            self.__dict__ = loss_accuracy_dictionary



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import cv2
import numpy as np
import multiprocessing
from joblib import Parallel, delayed
from confapp import conf

# from idtrackerai.utils.py_utils import *
from idtrackerai.utils.py_utils import set_mkl_to_single_thread, set_mkl_to_multi_thread

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.segmentation_utils")


"""
The utilities to segment and extract the blob information
"""


def sum_frames_for_bkg_per_episode_in_single_file_video(starting_frame,
                                                        ending_frame,
                                                        video_path, bkg):
    """Computes the sum of frames (1 every 100 frames) for a particular episode of
    the video when the video is a single file.

    Parameters
    ----------
    starting_frame : int
        First frame of the episode
    ending_frame : int
        Last frame of the episode
    video_path : string
        Path to the single file of the video
    bkg : nd.array
        Zeros array with same width and height as the frame of the video.

    Returns
    -------
    bkg : nd.array
        Array with same width and height as the frame of the video. Contains the
        sum of (ending_frame - starting_frame) / 100 frames for the given episode
    number_of_frames_for_bkg_in_episode : int
        Number of frames used to compute the background in the current episode
    """
    cap = cv2.VideoCapture(video_path)
    logger.debug('Adding from starting frame %i to background' %starting_frame)
    number_of_frames_for_bkg_in_episode = 0
    frameInds = range(starting_frame,ending_frame, conf.BACKGROUND_SUBTRACTION_PERIOD)
    for ind in frameInds:
        logger.debug('Frame %i' %ind)
        cap.set(1,ind)
        ret, frameBkg = cap.read()
        if conf.SIGMA_GAUSSIAN_BLURRING is not None:
            frameBkg = cv2.GaussianBlur(frameBkg, (0, 0), conf.SIGMA_GAUSSIAN_BLURRING)
        if ret:
            gray = cv2.cvtColor(frameBkg, cv2.COLOR_BGR2GRAY)
            gray = np.true_divide(gray,np.mean(gray))
            bkg = bkg + gray
            number_of_frames_for_bkg_in_episode += 1

    cap.release()
    return bkg, number_of_frames_for_bkg_in_episode


def sum_frames_for_bkg_per_episode_in_multiple_files_video(video_path, bkg):
    """Computes the sum of frames (1 every 100 frames) for a particular episode of
    the video when the video is splitted in several files

    Parameters
    ----------
    video_path : string
        Path to the file of the episode to be added to the background
    bkg : nd.array
        Zeros array with same width and height as the frame of the video.

    Returns
    -------
    bkg : nd.array
        Array with same width and height as the frame of the video. Contains the
        sum of (ending_frame - starting_frame) / 100 frames for the given episode
    number_of_frames_for_bkg_in_episode : int
        Number of frames used to compute the background in the current episode
    """
    logger.debug('Adding video %s to background' % video_path)
    cap = cv2.VideoCapture(video_path)
    numFrame = int(cap.get(7))
    number_of_frames_for_bkg_in_episode = 0
    frameInds = range(0,numFrame, conf.BACKGROUND_SUBTRACTION_PERIOD)
    for ind in frameInds:
        cap.set(1,ind)
        ret, frameBkg = cap.read()
        if conf.SIGMA_GAUSSIAN_BLURRING is not None:
            frameBkg = cv2.GaussianBlur(frameBkg, (0, 0), conf.SIGMA_GAUSSIAN_BLURRING)
        if ret:
            gray = cv2.cvtColor(frameBkg, cv2.COLOR_BGR2GRAY)
            gray = np.true_divide(gray,np.mean(gray))
            bkg = bkg + gray
            number_of_frames_for_bkg_in_episode += 1

    return bkg, number_of_frames_for_bkg_in_episode


def cumpute_background(video):
    """Computes a model of the background by averaging multiple frames of the video.
    In particular 1 every 100 frames is used for the computation.

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading

    Returns
    -------
    bkg : nd.array
        Array with the model of the background.

    See Also
    --------
    sum_frames_for_bkg_per_episode_in_single_file_video
    sum_frames_for_bkg_per_episode_in_multiple_files_video
    """
    # This holds even if we have not selected a ROI because then the ROI is
    # initialized as the full frame
    bkg = np.zeros((video.original_height, video.original_width))

    set_mkl_to_single_thread()
    if video.paths_to_video_segments is None: # one single file
        logger.debug('one single video, computing bkg in parallel from single video')
        output = Parallel(n_jobs=NUMBER_OF_JOBS_FOR_BACKGROUND_SUBTRACTION)(delayed(
                    sum_frames_for_bkg_per_episode_in_single_file_video)(
                    starting_frame, ending_frame, video.video_path, bkg)
                    for (starting_frame, ending_frame) in video.episodes_start_end)
        logger.debug('Finished parallel loop for bkg subtraction')
    else: # multiple video files
        logger.debug('multiple videos, computing bkg in parallel from every episode')
        output = Parallel(n_jobs=NUMBER_OF_JOBS_FOR_BACKGROUND_SUBTRACTION)(delayed(
                    sum_frames_for_bkg_per_episode_in_multiple_files_video)(
                    videoPath,bkg) for videoPath in video.paths_to_video_segments)
        logger.debug('Finished parallel loop for bkg subtraction')
    set_mkl_to_multi_thread()

    partialBkg = [bkg for (bkg,_) in output]
    totNumFrame = np.sum([numFrame for (_,numFrame) in output])
    bkg = np.sum(np.asarray(partialBkg),axis=0)
    bkg = np.true_divide(bkg, totNumFrame)
    return bkg.astype('float32')

def segment_frame(frame, min_threshold, max_threshold, bkg, ROI, useBkg):
    """Applies the intensity thresholds (`min_threshold` and `max_threshold`) and the
    mask (`ROI`) to a given frame. If `useBkg` is True, the background subtraction
    operation is applied before thresholding with the given `bkg`.

    Parameters
    ----------
    frame : nd.array
        Frame to be segmented
    min_threshold : int
        Minimum intensity threshold for the segmentation (value from 0 to 255)
    max_threshold : int
        Maximum intensity threshold for the segmentation (value from 0 to 255)
    bkg : nd.array
        Background model to be used in the background subtraction operation
    ROI : nd.array
        Mask to be applied after thresholding. Ones in the array are pixels to be
        considered, zeros are pixels to be discarded.
    useBkg : bool
        Flag indicating whether background subtraction must be performed or not

    Returns
    -------
    frame_segmented_and_masked : nd.array
        Frame with zeros and ones after applying the thresholding and the mask.
        Pixels with value 1 are valid pixels given the thresholds and the mask.
    """
    if useBkg:
        frame = cv2.absdiff(bkg,frame) #only step where frame normalization is important, because the background is normalised
        p99 = np.percentile(frame, 99.95)*1.001
        frame = np.clip(255 - frame * (255.0/p99), 0, 255)
        frame_segmented = cv2.inRange(frame, min_threshold, max_threshold) #output: 255 in range, else 0
    elif not useBkg:
        p99 = np.percentile(frame, 99.95)*1.001
        frame_segmented = cv2.inRange(np.clip(frame * (255.0/p99), 0, 255), min_threshold, max_threshold) #output: 255 in range, else 0
    frame_segmented_and_masked = cv2.bitwise_and(frame_segmented,frame_segmented, mask=ROI) #Applying the mask
    return frame_segmented_and_masked


def filter_contours_by_area(contours, min_area, max_area):
    """Filters out contours which number of pixels is smaller than `min_area` or
    greater than `max_area`

    Parameters
    ----------
    contours : list
        List of OpenCV contours
    min_area : int
        Minimum number of pixels for a contour to be acceptable
    max_area : int
        Maximum number of pixels for a contours to be acceptable

    Returns
    -------
    good_contours : list
        List of OpenCV contours that fulfill both area thresholds
    """

    good_contours = []
    for contour in contours:
        area = cv2.contourArea(contour)
        if area > min_area and area < max_area:
            good_contours.append(contour)
    return good_contours


def cnt2BoundingBox(cnt,bounding_box):
    """Transforms the coordinates of the contour in the full frame to the the
    bounding box of the blob.

    Parameters
    ----------
    cnt : list
        List of the coordinates that defines the contour of the blob in the full
        frame of the video
    bounding_box : tuple
        Tuple with the coordinates of the bounding box (x, y),(x + w, y + h))


    Returns
    -------
    contour_in_bounding_box : nd.array
        Array with the pairs of coordinates of the contour in the bounding box
    """
    return cnt - np.asarray([bounding_box[0][0],bounding_box[0][1]])


def get_bounding_box(cnt, width, height, crossing_detector = False):
    """Computes the bounding box of a given contour with an extra margin of 45
    pixels. If the function is called with the crossing_detector set to True the
    margin of the bounding box is set to 55.

    Parameters
    ----------
    cnt : list
        List of the coordinates that defines the contour of the blob in the full
        frame of the video
    width : int
        Width of the video frame
    height : int
        Height of the video frame
    crossing_detector : bool
        Flag to indicate whether the function is being called from the crossing_detector module

    Returns
    -------
    bounding_box : tuple
        Tuple with the coordinates of the bounding box (x, y),(x + w, y + h))
    original_diagonal : int
        Diagonal of the original bounding box computed with OpenCv that serves as
        an estimate for the body length of the animal.
    """
    x,y,w,h = cv2.boundingRect(cnt)
    original_diagonal = int(np.ceil(np.sqrt(w**2 + h**2)))
    n = 45 if not crossing_detector else 55
    if x - n > 0: # We only expand the
        x = x - n
    else:
        x = 0
    if y - n > 0:
        y = y - n
    else:
        y = 0
    if x + w + 2*n < width:
        w = w + 2*n
    else:
        w = width - x
    if y + h + 2*n < height:
        h = h + 2*n
    else:
        h = height - y
    return ((x, y),(x + w, y + h)), original_diagonal


def getCentroid(cnt):
    """Computes the centroid of the contour

    Parameters
    ----------
    cnt : list
        List of the coordinates that defines the contour of the blob in the full
        frame of the video

    Returns
    -------
    centroid : tuple
        (x,y) coordinates of the center of mass of the contour.
    """
    M = cv2.moments(cnt)
    x = M['m10']/M['m00']
    y = M['m01']/M['m00']
    return (x,y)


def get_pixels(cnt, width, height):
    """Gets the coordinates list of the pixels inside the contour

    Parameters
    ----------
    cnt : list
        List of the coordinates that defines the contour of the blob in a give
        width and height (it can either be the video width and heigh or the
        bounding box width and height)
    width : int
        Width of the frame
    height : int
        Height of the frame

    Returns
    -------
    pixels_coordinates_list : list
        List of the coordinates of the pixels in a given width and height
    """
    cimg = np.zeros((height, width))
    cv2.drawContours(cimg, [cnt], -1, color=255, thickness = -1)
    pts = np.where(cimg == 255)
    return np.asarray(list(zip(pts[0],pts[1])))


def get_bounding_box_image(frame, cnt, save_pixels, save_segmentation_image):
    """Computes the `bounding_box_image`from a given frame and contour. It also
    returns the coordinates of the `bounding_box`, the ravelled `pixels` inside of
    the contour and the diagonal of the `bounding_box` as an `estimated_body_length`

    Parameters
    ----------
    frame : nd.array
        frame from where to extract the `bounding_box_image`
    cnt : list
        List of the coordinates that defines the contour of the blob in the full
        frame of the video

    Returns
    -------
    bounding_box : tuple
        Tuple with the coordinates of the bounding box (x, y),(x + w, y + h))
    bounding_box_image : nd.array
        Part of the `frame` defined by the coordinates in `bounding_box`
    pixels_in_full_frame_ravelled : list
        List of ravelled pixels coordinates inside of the given contour
    estimated_body_length : int
        Estimated length of the contour in pixels.

    See Also
    --------
    get_bounding_box
    cnt2BoundingBox
    get_pixels
    """
    height = frame.shape[0]
    width = frame.shape[1]
    bounding_box, estimated_body_length = get_bounding_box(cnt, width, height) # the estimated body length is the diagonal of the original bounding_box
    if save_segmentation_image == 'RAM' or save_segmentation_image == 'DISK':
        bounding_box_image = frame[bounding_box[0][1]:bounding_box[1][1],
                                bounding_box[0][0]:bounding_box[1][0]]
    elif save_segmentation_image == 'NONE':
        bounding_box_image = None
    contour_in_bounding_box = cnt2BoundingBox(cnt, bounding_box)
    if save_pixels == 'RAM' or save_pixels == 'DISK':
        pixels_in_bounding_box = get_pixels(contour_in_bounding_box,
                                np.abs(bounding_box[0][0] - bounding_box[1][0]),
                                np.abs(bounding_box[0][1] - bounding_box[1][1]))
        pixels_in_full_frame = pixels_in_bounding_box + \
                                np.asarray([bounding_box[0][1], bounding_box[0][0]])
        pixels_in_full_frame_ravelled = np.ravel_multi_index(
                                        [pixels_in_full_frame[:,0], pixels_in_full_frame[:,1]],
                                        (height,width))
    elif save_pixels == 'NONE':
        pixels_in_full_frame_ravelled = None

    return bounding_box, bounding_box_image, pixels_in_full_frame_ravelled, estimated_body_length


def get_blobs_information_per_frame(frame, contours, save_pixels,
                                    save_segmentation_image):
    """Computes a set of properties for all the `contours` in a given frame.

    Parameters
    ----------
    frame : nd.array
        Frame from where to extract the `bounding_box_image` of every contour
    contours : list
        List of OpenCV contours for which to compute the set of properties

    Returns
    -------
    bounding_boxes : list
        List with the `bounding_box` for every contour in `contours`
    bounding_box_images : list
        List with the `bounding_box_image` for every contour in `contours`
    centroids : list
        List with the `centroid` for every contour in `contours`
    areas : list
        List with the `area` in pixels for every contour in `contours`
    pixels : list
        List with the `pixels` for every contour in `contours`
    estimated_body_lengths : list
        List with the `estimated_body_length` for every contour in `contours`

    See Also
    --------
    get_bounding_box_image
    getCentroid
    get_pixels
    """
    bounding_boxes = []
    bounding_box_images = []
    centroids = []
    areas = []
    pixels = []
    estimated_body_lengths = []

    for i, cnt in enumerate(contours):
        bounding_box, \
        bounding_box_image, \
        pixels_in_full_frame_ravelled, \
        estimated_body_length = get_bounding_box_image(frame, cnt, save_pixels,
                                                       save_segmentation_image)
        #bounding boxes
        bounding_boxes.append(bounding_box)
        # bounding_box_images
        bounding_box_images.append(bounding_box_image)
        # centroids
        centroids.append(getCentroid(cnt))
        areas.append(cv2.contourArea(cnt))
        # pixels lists
        pixels.append(pixels_in_full_frame_ravelled)
        # estimated body lengths list
        estimated_body_lengths.append(estimated_body_length)

    return bounding_boxes, bounding_box_images, centroids, areas, \
        pixels, estimated_body_lengths


def blob_extractor(segmented_frame, frame, min_area, max_area, save_pixels='DISK',
                   save_segmentation_image='DISK'):
    """Given a `segmented_frame` it extracts the blobs with area greater than
    `min_area` and smaller than `max_area` and it computes a set of relevant
    properties for every blob.

    Parameters
    ----------
    segmented_frame : nd.array
        Frame with zeros and ones where ones are valid pixels.
    frame : nd.array
        Frame from where to extract the `bounding_box_image` of every blob
    min_area : int
        Minimum number of blobs above which a blob is considered to be valid
    max_area : int
        Maximum number of blobs below which a blob is considered to be valid

    Returns
    -------
    bounding_boxes : list
        List with the `bounding_box` for every contour in `contours`
    bounding_box_images : list
        List with the `bounding_box_image` for every contour in `contours`
    centroids : list
        List with the `centroid` for every contour in `contours`
    areas : list
        List with the `area` in pixels for every contour in `contours`
    pixels : list
        List with the `pixels` for every contour in `contours`
    good_contours_in_full_frame:
        List with the `contours` which area is in between `min_area` and `max_area`
    estimated_body_lengths : list
        List with the `estimated_body_length` for every contour in `contours`

    See Also
    --------
    filter_contours_by_area
    get_blobs_information_per_frame
    """
    _, contours, hierarchy = cv2.findContours(segmented_frame,
                                              cv2.RETR_TREE,
                                              cv2.CHAIN_APPROX_NONE)
    # Filter contours by size
    good_contours_in_full_frame = filter_contours_by_area(contours,
                                                          min_area,
                                                          max_area)
    # get contours properties
    bounding_boxes, bounding_box_images, \
        centroids, areas, pixels, \
        estimated_body_lengths = \
        get_blobs_information_per_frame(frame, good_contours_in_full_frame,
                                        save_pixels, save_segmentation_image)

    return bounding_boxes, bounding_box_images, centroids, areas, pixels, \
        good_contours_in_full_frame, estimated_body_lengths



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
import h5py
import tensorflow as tf

from confapp import conf

from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.network.cnn_architectures import cnn_model_crossing_detector
from idtrackerai.network.crossings_detector_model.get_crossings_data_set import CrossingDataset
from idtrackerai.network.crossings_detector_model.network_params_crossings import NetworkParams_crossings
from idtrackerai.network.crossings_detector_model.crossings_detector_model import ConvNetwork_crossings
from idtrackerai.network.crossings_detector_model.train_crossings_detector import TrainDeepCrossing
from idtrackerai.network.crossings_detector_model.get_predictions_crossings import GetPredictionCrossigns

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.crossing_detector")

# FIXME: This function returns either a TrainDeepCrossing object or a
# ListOfBlobs. Not good practice.


def detect_crossings(list_of_blobs,
                     video,
                     model_area,
                     use_network=True,
                     return_store_objects=False,
                     plot_flag=False):
    """Classify all blobs in the video as bing crossings or individuals.

    Parameters
    ----------
    list_of_blobs : <ListOfBlobs object>
        Collection of the Blob objects extracted from the video
    video :  <Video object>
        Object containing all the parameters of the video.
    model_area : function
        Model of the area of a single individual
    use_network : bool
        If True the Deep Crossing Detector is used to distinguish between
        individuals and crossings images. Otherwise only the model area is applied
    return_store_objects : bool
        If True the instantiations of the class :class:`.Store_Accuracy_and_Loss`
        are returned by the function
    plot_flag : bool
        If True a figure representing the values of the loss function, accuracy
        and accuracy per class for both the training and validation set.

    Returns
    -------

    trainer or list_of_blobs : TrainDeepCrossing or ListOfBlobs()
    """
    logger.info("Discriminating blobs representing individuals from blobs associated to crossings")

    image_shape = video.identification_image_size[0]
    with h5py.File(video.identification_images_file_path, 'w') as f:
        f.create_dataset("identification_images", ((0, image_shape, image_shape)),
                         chunks=(1, image_shape, image_shape),
                         maxshape=(video.number_of_animals * video.number_of_frames * 5,
                                   image_shape, image_shape))
        f.attrs['number_of_animals'] = video.number_of_animals
        f.attrs['video_path'] = video.video_path

    list_of_blobs.apply_model_area_to_video(video, model_area, video.identification_image_size[0], video.number_of_animals)

    if use_network:
        tf.reset_default_graph()
        video.create_crossings_detector_folder()
        logger.info("Get individual and crossing images labelled data")
        training_set = CrossingDataset(list_of_blobs.blobs_in_video,
                                       video,
                                       scope='training')
        if not training_set.there_are_crossings:
            logger.debug("There are not enough crossings to train the crossing detector")
            video._there_are_crossings = False
            return list_of_blobs
        else:
            training_set.get_data(sampling_ratio_start=0,
                                  sampling_ratio_end=.9)
            validation_set = CrossingDataset(list_of_blobs.blobs_in_video,
                                             video,
                                             scope='validation',
                                             crossings=training_set.crossing_blobs,
                                             individual_blobs=training_set.individual_blobs,
                                            image_size=training_set.image_size)
            validation_set.get_data(sampling_ratio_start=.9, sampling_ratio_end=1.)
            logger.info("Start crossing detector training")
            logger.info("Crossing detector training finished")
            crossing_image_size = training_set.image_size
            crossing_image_shape = training_set.images.shape[1:]
            logger.info("crossing image shape %s" %str(crossing_image_shape))
            crossings_detector_network_params = \
                NetworkParams_crossings(number_of_classes=2,
                                        learning_rate=conf.LEARNING_RATE_DCD,
                                        architecture=cnn_model_crossing_detector,
                                        keep_prob=conf.KEEP_PROB_IDCNN_PRETRAINING,
                                        save_folder=video._crossings_detector_folder,
                                        image_size=crossing_image_shape)
            net = ConvNetwork_crossings(crossings_detector_network_params)
            trainer = TrainDeepCrossing(net, training_set, validation_set,
                                        num_epochs=95,
                                        plot_flag=plot_flag,
                                        return_store_objects=return_store_objects)
        if not trainer.model_diverged:
            logger.debug("crossing image size %s" %str(crossing_image_size))
            video.crossing_image_shape = crossing_image_shape
            video.crossing_image_size = crossing_image_size
            video.save()
            # ind_blob_images = np.asarray(validation_set.generate_individual_blobs_images())
            # crossing_images = validation_set.generate_crossing_images()
            validation_set = None
            training_set = None
            logger.debug("Freeing memory. Validation and training crossings sets deleted")
            test_set = CrossingDataset(list_of_blobs.blobs_in_video, video,
                                       scope='test',
                                       image_size=video.crossing_image_size)
            logger.debug("Classify individuals and crossings")
            crossings_predictor = GetPredictionCrossigns(net)
            predictions = crossings_predictor.get_all_predictions(test_set)
            for blob, prediction in zip(test_set.test, predictions):
                if prediction == 1:
                    blob._is_a_crossing = True
                    blob._is_an_individual = False
                else:
                    blob._is_a_crossing = False
                    blob._is_an_individual = True
                    blob.set_image_for_identification(video)
            # [(setattr(blob, '_is_a_crossing', True), setattr(blob, '_is_an_individual', False)) if prediction == 1
            #     else (setattr(blob, '_is_a_crossing', False), setattr(blob, '_is_an_individual', True), blob.set_image_for_identification(video))
            #     for blob, prediction in zip(test_set.test, predictions)]
            logger.debug("Freeing memory. Test crossings set deleted")
            test_set = None
            if return_store_objects:
                return trainer
    if video.number_of_animals == 1:
        video._there_are_crossings = False
        return list_of_blobs



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import numpy as np
import matplotlib.pyplot as plt

from idtrackerai.utils.py_utils import  get_spaced_colors_util

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.store_accuracy_and_loss")

class Store_Accuracy_and_Loss(object):
    """Store the loss, accuracy and individual accuracy values computed during
    training and validation

    Parameters
    ----------
    _path_to_accuracy_error_data :
        Path to save the lists :attr:`loss`, :attr:`accuracy`, :attr:`individual_accuracy`
    name : string
        'Training' or 'Validation'
    loss : list
        List with the values of the loss
    accuracy : list
        List with the values of the accuracy
    individual_accuracy : list
        List with the values of the individual accuracies
    number_of_epochs_completed :
        Number of epochs completed
    scope : string
        'Training' if the class is instantiated during the training of the accumulation.
        'Pretraining' if the class is instantiated during the training
    """
    def __init__(self, network, name, scope = None):
        self._path_to_accuracy_error_data = network.params.save_folder
        self.name = name
        self.loss = []
        self.accuracy = []
        self.individual_accuracy = []
        self.number_of_epochs_completed = []
        self.scope = scope
        self.load()

    def append_data(self, loss_value, accuracy_value, individual_accuracy_value):
        """Appends the `loss_value`, `accuracy_value` and `individual_accuracy_value`
        to their correspoding lists
        """
        self.loss.append(loss_value)
        self.accuracy.append(accuracy_value)
        self.individual_accuracy.append(individual_accuracy_value)

    def plot(self, axes_handles = None, index = 0, color = 'r', canvas_from_GUI = None, legend_font_color = None):
        """Plots the accuracy and the individual accuracies for every epoch
        """
        if canvas_from_GUI is not None:
            ax1 = axes_handles[0]
            ax2 = axes_handles[1]
            ax1.plot(range(len(self.loss)),self.accuracy, color,label=self.name)
            width = 0.35
            numAnimals = len(self.individual_accuracy[-1])
            if self.name == 'training':
                ind = np.arange(numAnimals) + 1 - width
            elif self.name == 'validation':
                ind = np.arange(numAnimals) + 1
            ax1.set_xlabel('Epochs')
            ax2.bar(ind, self.individual_accuracy[-1], width, color=color, alpha=0.4,label=self.name)
            ax2.set_xlabel('Individual')
            ax2.set_ylabel('Individual \naccuracy')
            if index == 0:
                legend = ax1.legend()
                if legend_font_color is not None:
                    for text in legend.get_texts():
                        text.set_color(legend_font_color)
                ax1.set_ylabel('Accuracy')
        else:
            ax1 = axes_handles[0]
            ax2 = axes_handles[1]
            ax3 = axes_handles[2]
            ax1.plot(range(index,len(self.loss)), self.loss[index:], color,label=self.name)
            ax2.plot(range(index,len(self.loss)),self.accuracy[index:], color,label=self.name)
            ax2.set_xlabel('Epochs')
            width = 0.35
            numAnimals = len(self.individual_accuracy[-1])
            if self.name == 'training':
                ind = np.arange(numAnimals)+1-width
            elif self.name == 'validation':
                ind = np.arange(numAnimals)+1
            ax3.bar(ind, self.individual_accuracy[-1], width, color=color, alpha=0.4,label=self.name)
            ax3.set_xlabel('Individual')
            ax3.set_ylabel('Individual \naccuracy')
            if index == 0:
                legend = ax1.legend()
                if legend_font_color is not None:
                    for text in legend.get_texts():
                        text.set_color(legend_font_color)
                ax1.set_ylabel('Loss')
                ax2.set_ylabel('Accuracy')
                ax2.set_xlabel('Epochs')
        if canvas_from_GUI is None:
            plt.draw()
            plt.pause(1e-8)
        else:
            canvas_from_GUI.draw()

    def plot_global_fragments(self, ax_handles, video, fragments, black = False, canvas_from_GUI = None):
        """Plots the global fragments used for training until the current epoch
        """
        import matplotlib.patches as patches
        if canvas_from_GUI is not None:
            ax4 = ax_handles[2]
            ax4.cla()
        else:
            ax4 = ax_handles[3]
            ax4.cla()
        colors = get_spaced_colors_util(video.number_of_animals, norm=True, black=black)
        attribute_to_check = 'used_for_training' if self.scope == 'training' else 'used_for_pretraining'

        for fragment in fragments:
            if getattr(fragment, attribute_to_check) and fragment.is_an_individual:
                if fragment.final_identities is not None and len(fragment.final_identities)==1:
                    blob_index = fragment.final_identities[0] - 1
                elif self.scope == 'pretraining':
                    blob_index = fragment.blob_hierarchy_in_starting_frame
                else:
                    blob_index = fragment.temporary_id
                (start, end) = fragment.start_end
                ax4.add_patch(
                    patches.Rectangle(
                        (start, blob_index - 0.5),   # (x,y)
                        end - start,  # width
                        1.,          # height
                        fill=True,
                        edgecolor=None,
                        facecolor=colors[fragment.temporary_id if attribute_to_check == 'used_for_training' else fragment._temporary_id_for_pretraining],
                        alpha = 1.
                    )
                )
        ax4.axis('tight')
        ax4.set_xlabel('Frame number')
        ax4.set_ylabel('Blob index')
        if not hasattr(video, 'maximum_number_of_blobs') or video.maximum_number_of_blobs == 0:
            max_y = video.number_of_animals
        else:
            max_y = video.maximum_number_of_blobs
        ax4.set_yticks(range(0, max_y, 3))
        ax4.set_yticklabels(range(1, max_y + 1, 3))
        ax4.set_xlim([0., video.number_of_frames])
        ax4.set_ylim([-.5, max_y + .5 - 1])

    def save(self, number_of_epochs_completed):
        """Saves the values stored
        """
        self.number_of_epochs_completed.append(number_of_epochs_completed)
        np.save(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy'), self.__dict__)

    def load(self):
        """Load the values stored in case there are any saved
        """
        if os.path.isfile(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy')):
            loss_accuracy_dictionary = np.load(os.path.join(self._path_to_accuracy_error_data, self.name + '_loss_acc_dict.npy'), allow_pickle=True).item()
            self.__dict__ = loss_accuracy_dictionary
            if not hasattr(self, 'number_of_epochs_completed'):
                self.number_of_epochs_completed = []
